# -*- coding: utf-8 -*-
__author__ = 'isparks'

import uuid
from xml.etree import cElementTree as ET
from datetime import datetime
from string import ascii_letters
from rwslib.builder_constants import *

"""
builders.py provides convenience classes for building ODM documents for clinical data and metadata post messages.
"""

# -----------------------------------------------------------------------------------------------------------------------
# Constants

VALID_ID_CHARS = ascii_letters + '_'


# -----------------------------------------------------------------------------------------------------------------------
# Utilities


def now_to_iso8601():
    """Returns NOW date/time as a UTC date/time formated as iso8601 string"""
    utc_date = datetime.utcnow()
    return dt_to_iso8601(utc_date)


def dt_to_iso8601(dt):
    """Turn a datetime into an ISO8601 formatted string"""
    return dt.strftime("%Y-%m-%dT%H:%M:%S")


def bool_to_yes_no(val):
    """Convert True/False to Yes/No"""
    return 'Yes' if val else 'No'


def bool_to_true_false(val):
    """Convert True/False to TRUE / FALSE"""
    return 'TRUE' if val else 'FALSE'


def indent(elem, level=0):
    """Indent a elementree structure"""
    i = "\n" + level * "  "
    if len(elem) > 0:
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def make_element(builder, tag, content):
    """Make an element with this tag and text content"""
    builder.start(tag, {})
    builder.data(content)  # Must be UTF-8 encoded
    builder.end(tag)


# -----------------------------------------------------------------------------------------------------------------------
# Classes

class ODMElement(object):
    """Base class for ODM XML element classes"""

    def __call__(self, *args):
        """Collect all children passed in call"""
        for child in args:
            self << child
        return self

    def __lshift__(self, other):
        """__lshift__ should be overridden in descendant classes to accept child elements and incorporate them.
           By default takes no child elements
        """
        raise ValueError("%s takes no child elements" % self.__class__.__name__)

    def add(self, *args):
        """Like call but adds a set of args"""
        for child in args:
            self << child
        return self

    def set_single_attribute(self, other, trigger_klass, property_name):
        """Used to set guard the setting of an attribute which is singular and can't be set twice"""

        if isinstance(other, trigger_klass):

            # Check property exists
            if not hasattr(self, property_name):
                raise AttributeError("%s has no property %s" % (self.__class__.__name__, property_name))

            if getattr(self, property_name) is None:
                setattr(self, property_name, other)
            else:
                raise ValueError(
                    '%s already has a %s element set.' % (self.__class__.__name__, other.__class__.__name__,))

    def set_list_attribute(self, other, trigger_klass, property_name):
        """Used to set guard the setting of a list attribute, ensuring the same element is not added twice."""
        # Check property exists
        if isinstance(other, trigger_klass):

            if not hasattr(self, property_name):
                raise AttributeError("%s has no property %s" % (self.__class__.__name__, property_name))

            val = getattr(self, property_name, [])
            if other in val:
                raise ValueError("%s already exists in %s" % (other.__class__.__name__, self.__class__.__name__))
            else:
                val.append(other)
                setattr(self, property_name, val)


class UserRef(ODMElement):
    def __init__(self, oid):
        self.oid = oid

    def build(self, builder):
        builder.start("UserRef", dict(UserOID=self.oid))
        builder.end("UserRef")


class LocationRef(ODMElement):
    def __init__(self, oid):
        self.oid = oid

    def build(self, builder):
        builder.start("LocationRef", dict(LocationOID=self.oid))
        builder.end("LocationRef")


class ReasonForChange(ODMElement):
    def __init__(self, reason):
        self.reason = reason

    def build(self, builder):
        builder.start("ReasonForChange", {})
        builder.data(self.reason)
        builder.end("ReasonForChange")


class DateTimeStamp(ODMElement):
    def __init__(self, date_time):
        self.date_time = date_time

    def build(self, builder):
        builder.start("DateTimeStamp", {})
        if isinstance(self.date_time, datetime):
            builder.data(dt_to_iso8601(self.date_time))
        else:
            builder.data(self.date_time)
        builder.end("DateTimeStamp")


class AuditRecord(ODMElement):
    """AuditRecord is supported only by ItemData in Rave"""
    EDIT_MONITORING = 'Monitoring'
    EDIT_DATA_MANAGEMENT = 'DataManagement'
    EDIT_DB_AUDIT = 'DBAudit'
    EDIT_POINTS = [EDIT_MONITORING, EDIT_DATA_MANAGEMENT, EDIT_DB_AUDIT]

    def __init__(self, edit_point=None, used_imputation_method=None, identifier=None, include_file_oid=None):
        self._edit_point = None
        self.edit_point = edit_point
        self.used_imputation_method = used_imputation_method
        self._id = None
        self.id = identifier
        self.include_file_oid = include_file_oid
        self.user_ref = None
        self.location_ref = None
        self.reason_for_change = None
        self.date_time_stamp = None

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if value not in [None, ''] and str(value).strip() != '':
            val = str(value).strip()[0]
            if val not in VALID_ID_CHARS:
                raise AttributeError('%s id cannot start with "%s" character' % (self.__class__.__name__, val,))
        self._id = value

    @property
    def edit_point(self):
        return self._edit_point

    @edit_point.setter
    def edit_point(self, value):
        if value is not None:
            if value not in self.EDIT_POINTS:
                raise AttributeError('%s edit_point must be one of %s not %s' % (
                    self.__class__.__name__, ','.join(self.EDIT_POINTS), value,))
        self._edit_point = value

    def build(self, builder):
        params = {}

        if self.edit_point is not None:
            params["EditPoint"] = self.edit_point

        if self.used_imputation_method is not None:
            params['UsedImputationMethod'] = bool_to_yes_no(self.used_imputation_method)

        if self.id is not None:
            params['ID'] = str(self.id)

        if self.include_file_oid is not None:
            params['mdsol:IncludeFileOID'] = bool_to_yes_no(self.include_file_oid)

        builder.start("AuditRecord", params)
        if self.user_ref is None:
            raise ValueError("User Reference not set.")
        self.user_ref.build(builder)

        if self.location_ref is None:
            raise ValueError("Location Reference not set.")
        self.location_ref.build(builder)

        if self.date_time_stamp is None:
            raise ValueError("DateTime not set.")

        self.date_time_stamp.build(builder)

        # Optional
        if self.reason_for_change is not None:
            self.reason_for_change.build(builder)

        builder.end("AuditRecord")

    def __lshift__(self, other):
        if not isinstance(other, (UserRef, LocationRef, DateTimeStamp, ReasonForChange,)):
            raise ValueError("AuditRecord cannot accept a child element of type %s" % other.__class__.__name__)

        # Order is important, apparently
        self.set_single_attribute(other, UserRef, 'user_ref')
        self.set_single_attribute(other, LocationRef, 'location_ref')
        self.set_single_attribute(other, DateTimeStamp, 'date_time_stamp')
        self.set_single_attribute(other, ReasonForChange, 'reason_for_change')
        return other


class TransactionalElement(ODMElement):
    """Models an ODM Element that is allowed a transaction type. Different elements have different
       allowed transaction types"""
    ALLOWED_TRANSACTION_TYPES = []

    def __init__(self, transaction_type):
        self._transaction_type = None
        self.transaction_type = transaction_type

    @property
    def transaction_type(self):
        return self._transaction_type

    @transaction_type.setter
    def transaction_type(self, value):
        if value is not None:
            if value not in self.ALLOWED_TRANSACTION_TYPES:
                raise AttributeError('%s transaction_type element must be one of %s not %s' % (
                    self.__class__.__name__, ','.join(self.ALLOWED_TRANSACTION_TYPES), value,))
        self._transaction_type = value


class MdsolQuery(ODMElement):
    """MdsolQuery extension element for Queries at item level only"""

    def __init__(self, value=None, query_repeat_key=None, recipient=None, status=None, requires_response=None,
                 response=None):
        self.value = value
        self.query_repeat_key = query_repeat_key
        self.recipient = recipient
        self._status = None
        self.status = status
        self.requires_response = requires_response
        self.response = response

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        if value is not None:
            if not isinstance(value, QueryStatusType):
                raise AttributeError("%s action type is invalid in mdsol:Query." % (value,))
        self._status = value

    def build(self, builder):
        params = {}

        if self.value is not None:
            params['Value'] = str(self.value)

        if self.query_repeat_key is not None:
            params['QueryRepeatKey'] = str(self.query_repeat_key)

        if self.recipient is not None:
            params['Recipient'] = str(self.recipient)

        if self.status is not None:
            params['Status'] = self.status.value

        if self.requires_response is not None:
            params['RequiresResponse'] = bool_to_yes_no(self.requires_response)

        # When closing a query
        if self.response is not None:
            params['Response'] = str(self.response)

        builder.start("mdsol:Query", params)
        builder.end("mdsol:Query")


class ItemData(TransactionalElement):
    """Models the ODM ItemData object"""
    ALLOWED_TRANSACTION_TYPES = ['Insert', 'Update', 'Upsert', 'Context', 'Remove']

    def __init__(self, itemoid, value, specify_value=None, transaction_type=None, lock=None, freeze=None, verify=None):
        super(self.__class__, self).__init__(transaction_type)
        self.itemoid = itemoid
        self.value = value

        self.specify_value = specify_value
        self.lock = lock
        self.freeze = freeze
        self.verify = verify
        self.audit_record = None
        self.queries = []
        self.measurement_unit_ref = None

    def build(self, builder):
        """Build XML by appending to builder
           <ItemData ItemOID="MH_DT" Value="06 Jan 2009" TransactionType="Insert">
        """
        params = dict(ItemOID=self.itemoid)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.value in [None, '']:
            params['IsNull'] = 'Yes'
        else:
            params['Value'] = str(self.value)

        if self.specify_value is not None:
            params['mdsol:SpecifyValue'] = self.specify_value

        if self.lock is not None:
            params['mdsol:Lock'] = bool_to_yes_no(self.lock)

        if self.freeze is not None:
            params['mdsol:Freeze'] = bool_to_yes_no(self.freeze)

        if self.verify is not None:
            params['mdsol:Verify'] = bool_to_yes_no(self.verify)

        builder.start("ItemData", params)


        if self.audit_record is not None:
            self.audit_record.build(builder)

        # Measurement unit ref must be after audit record or RWS complains
        if self.measurement_unit_ref is not None:
            self.measurement_unit_ref.build(builder)

        for query in self.queries:
            query.build(builder)
        builder.end("ItemData")

    def __lshift__(self, other):
        if not isinstance(other, (MeasurementUnitRef, AuditRecord, MdsolQuery,)):
            raise ValueError("ItemData object can only receive MeasurementUnitRef, AuditRecord or MdsolQuery objects")
        self.set_single_attribute(other, MeasurementUnitRef, 'measurement_unit_ref')
        self.set_single_attribute(other, AuditRecord, 'audit_record')
        self.set_list_attribute(other, MdsolQuery, 'queries')
        return other


class ItemGroupData(TransactionalElement):
    """Models the ODM ItemGroupData object.
       Note no name for the ItemGroupData element is required. This is built automatically by the form.
    """
    ALLOWED_TRANSACTION_TYPES = ['Insert', 'Update', 'Upsert', 'Context']

    def __init__(self, transaction_type=None, item_group_repeat_key=None, whole_item_group=False):
        super(self.__class__, self).__init__(transaction_type)
        self.item_group_repeat_key = item_group_repeat_key
        self.whole_item_group = whole_item_group
        self.items = {}

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, ItemData):
            raise ValueError("ItemGroupData object can only receive ItemData object")

        if other.itemoid in self.items:
            raise ValueError("ItemGroupData object with that itemoid is already in the ItemGroupData object")

        self.items[other.itemoid] = other
        return other

    def build(self, builder, formname):
        """Build XML by appending to builder
        """
        params = dict(ItemGroupOID=formname)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.item_group_repeat_key is not None:
            params["ItemGroupRepeatKey"] = str(
                self.item_group_repeat_key)  # may be @context for transaction type upsert or context

        params["mdsol:Submission"] = "WholeItemGroup" if self.whole_item_group else "SpecifiedItemsOnly"

        builder.start("ItemGroupData", params)

        # Ask children
        for item in self.items.values():
            item.build(builder)
        builder.end("ItemGroupData")


class FormData(TransactionalElement):
    """Models the ODM FormData object"""
    ALLOWED_TRANSACTION_TYPES = ['Insert', 'Update']

    def __init__(self, formoid, transaction_type=None, form_repeat_key=None):
        super(self.__class__, self).__init__(transaction_type)
        self.formoid = formoid
        self.form_repeat_key = form_repeat_key
        self.itemgroups = []

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, ItemGroupData):
            raise ValueError("FormData object can only receive ItemGroupData object")
        self.set_list_attribute(other, ItemGroupData, 'itemgroups')
        return other

    def build(self, builder):
        """Build XML by appending to builder

        <FormData FormOID="MH" TransactionType="Update">
        """
        params = dict(FormOID=self.formoid)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.form_repeat_key is not None:
            params["FormRepeatKey"] = str(self.form_repeat_key)

        builder.start("FormData", params)

        # Ask children
        for itemgroup in self.itemgroups:
            itemgroup.build(builder, self.formoid)
        builder.end("FormData")


class StudyEventData(TransactionalElement):
    """Models the ODM StudyEventData object"""
    ALLOWED_TRANSACTION_TYPES = ['Insert', 'Update', 'Remove', 'Context']

    def __init__(self, study_event_oid, transaction_type="Update", study_event_repeat_key=None):
        super(self.__class__, self).__init__(transaction_type)
        self.study_event_oid = study_event_oid
        self.study_event_repeat_key = study_event_repeat_key
        self.forms = []

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, FormData):
            raise ValueError("StudyEventData object can only receive FormData object")
        self.set_list_attribute(other, FormData, 'forms')
        return other

    def build(self, builder):
        """Build XML by appending to builder

        <StudyEventData StudyEventOID="SCREENING" StudyEventRepeatKey="1" TransactionType="Update">
        """
        params = dict(StudyEventOID=self.study_event_oid)

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        if self.study_event_repeat_key is not None:
            params["StudyEventRepeatKey"] = self.study_event_repeat_key

        builder.start("StudyEventData", params)

        # Ask children
        for form in self.forms:
            form.build(builder)
        builder.end("StudyEventData")


class SubjectData(TransactionalElement):
    """Models the ODM SubjectData and ODM SiteRef objects"""
    ALLOWED_TRANSACTION_TYPES = ['Insert', 'Update', 'Upsert']

    def __init__(self, sitelocationoid, subject_key, subject_key_type="SubjectName", transaction_type="Update"):
        super(self.__class__, self).__init__(transaction_type)
        self.sitelocationoid = sitelocationoid
        self.subject_key = subject_key
        self.subject_key_type = subject_key_type
        self.study_events = []  # Can have collection

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, StudyEventData):
            raise ValueError("SubjectData object can only receive StudyEventData object")
        self.set_list_attribute(other, StudyEventData, 'study_events')
        return other

    def build(self, builder):
        """Build XML by appending to builder"""
        params = dict(SubjectKey=self.subject_key)
        params['mdsol:SubjectKeyType'] = self.subject_key_type

        if self.transaction_type is not None:
            params["TransactionType"] = self.transaction_type

        builder.start("SubjectData", params)

        builder.start("SiteRef", {'LocationOID': self.sitelocationoid})
        builder.end("SiteRef")

        # Ask children
        for event in self.study_events:
            event.build(builder)
        builder.end("SubjectData")


class ClinicalData(ODMElement):
    """Models the ODM ClinicalData object"""

    def __init__(self, projectname, environment):
        self.projectname = projectname
        self.environment = environment
        self.subject_data = None

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, SubjectData):
            raise ValueError("ClinicalData object can only receive SubjectData object")
        self.set_single_attribute(other, SubjectData, 'subject_data')
        return other

    def build(self, builder):
        """Build XML by appending to builder"""
        params = dict(MetaDataVersionOID='1',
                      StudyOID="%s (%s)" % (self.projectname, self.environment,),
                      )

        builder.start("ClinicalData", params)
        # Ask children
        if self.subject_data is not None:
            self.subject_data.build(builder)
        builder.end("ClinicalData")


class ODM(ODMElement):
    """Models the ODM object"""
    FILETYPE_TRANSACTIONAL = 'Transactional'
    FILETYPE_SNAPSHOT = 'Snapshot'

    def __init__(self, originator, description="", creationdatetime=now_to_iso8601(), fileoid=None, filetype=None):
        self.originator = originator  # Required
        self.description = description
        self.creationdatetime = creationdatetime
        # filetype will always be "Transactional"
        # ODM version will always be 1.3
        # Granularity="SingleSubject"
        # AsOfDateTime always OMITTED (it's optional)
        self.clinical_data = None
        self.study = None
        self.filetype = ODM.FILETYPE_TRANSACTIONAL if filetype is None else ODM.FILETYPE_SNAPSHOT

        # Create unique fileoid if none given
        self.fileoid = str(uuid.uuid4()) if fileoid is None else fileoid

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (ClinicalData, Study,)):
            raise ValueError("ODM object can only receive ClinicalData or Study object")
        self.set_single_attribute(other, ClinicalData, 'clinical_data')
        self.set_single_attribute(other, Study, 'study')

        return other

    def getroot(self):
        """Build XML object, return the root"""
        builder = ET.TreeBuilder()

        params = dict(ODMVersion="1.3",
                      FileType=self.filetype,
                      CreationDateTime=self.creationdatetime,
                      Originator=self.originator,
                      FileOID=self.fileoid,
                      xmlns="http://www.cdisc.org/ns/odm/v1.3",
                      )
        params['xmlns:mdsol'] = "http://www.mdsol.com/ns/odm/metadata"

        if self.description:
            params['Description'] = self.description

        builder.start("ODM", params)

        # Ask the children
        if self.study is not None:
            self.study.build(builder)

        if self.clinical_data is not None:
            self.clinical_data.build(builder)

        builder.end("ODM")
        return builder.close()

    def __str__(self):
        doc = self.getroot()
        indent(doc)
        header = '<?xml version="1.0" encoding="utf-8" ?>\n'
        return header + ET.tostring(doc, encoding='utf-8').decode('utf-8')


# -----------------------------------------------------------------------------------------------------------------------
# Metadata Objects


class GlobalVariables(ODMElement):
    """GlobalVariables Metadata element"""

    def __init__(self, protocol_name, name=None, description=''):
        """Name and description are not important. protocol_name maps to the Rave project name"""
        self.protocol_name = protocol_name
        self.name = name if name is not None else protocol_name
        self.description = description

    def build(self, builder):
        """Build XML by appending to builder"""
        builder.start("GlobalVariables", {})
        make_element(builder, 'StudyName', self.name)
        make_element(builder, 'StudyDescription', self.description)
        make_element(builder, 'ProtocolName', self.protocol_name)
        builder.end("GlobalVariables")


class TranslatedText(ODMElement):
    """Represents a language and a translated text for that language"""

    def __init__(self, text, lang=None):
        self.text = text
        self.lang = lang

    def build(self, builder):
        """Build XML by appending to builder"""
        params = {}
        if self.lang is not None:
            params['xml:lang'] = self.lang
        builder.start("TranslatedText", params)
        builder.data(self.text)
        builder.end("TranslatedText")


class Symbol(ODMElement):
    def __init__(self):
        self.translations = []

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, TranslatedText):
            raise ValueError("Symbol can only accept TranslatedText objects as children")
        self.set_list_attribute(other, TranslatedText, 'translations')

        return other

    def build(self, builder):
        """Build XML by appending to builder"""
        builder.start("Symbol", {})
        for child in self.translations:
            child.build(builder)
        builder.end("Symbol")


class MeasurementUnit(ODMElement):
    """A measurement unit"""

    def __init__(self,
                 oid,
                 name,
                 unit_dictionary_name=None,
                 constant_a=1,
                 constant_b=1,
                 constant_c=0,
                 constant_k=0,
                 standard_unit=False):
        self.symbols = []
        self.oid = oid
        self.name = name
        self.unit_dictionary_name = unit_dictionary_name
        self.constant_a = constant_a
        self.constant_b = constant_b
        self.constant_c = constant_c
        self.constant_k = constant_k
        self.standard_unit = standard_unit

    def build(self, builder):
        """Build XML by appending to builder"""

        params = dict(OID=self.oid,
                      Name=self.name)

        if self.unit_dictionary_name:
            params['mdsol:UnitDictionaryName'] = self.unit_dictionary_name

        for suffix in ['A', 'B', 'C', 'K']:
            val = getattr(self, 'constant_{0}'.format(suffix.lower()))
            params['mdsol:Constant{0}'.format(suffix)] = str(val)

        if self.standard_unit:
            params['mdsol:StandardUnit'] = 'Yes'

        builder.start("MeasurementUnit", params)
        for child in self.symbols:
            child.build(builder)
        builder.end("MeasurementUnit")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, Symbol):
            raise ValueError("MeasurementUnits object can only receive Symbol object")
        self.set_list_attribute(other, Symbol, 'symbols')

        return other


class BasicDefinitions(ODMElement):
    """Container for Measurement units"""

    def __init__(self):
        self.measurement_units = []

    def build(self, builder):
        """Build XML by appending to builder"""
        builder.start("BasicDefinitions", {})
        for child in self.measurement_units:
            child.build(builder)
        builder.end("BasicDefinitions")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, MeasurementUnit):
            raise ValueError("BasicDefinitions object can only receive MeasurementUnit object")
        self.measurement_units.append(other)
        return other


class StudyEventRef(ODMElement):
    def __init__(self, oid, order_number, mandatory):
        self.oid = oid
        self.order_number = order_number
        self.mandatory = mandatory

    def build(self, builder):
        """Build XML by appending to builder"""
        params = dict(StudyEventOID=self.oid,
                      OrderNumber=str(self.order_number),
                      Mandatory=bool_to_yes_no(self.mandatory))
        builder.start("StudyEventRef", params)
        builder.end("StudyEventRef")


class Protocol(ODMElement):
    """Protocol child of MetaDataVersion, holder of StudyEventRefs"""

    def __init__(self):
        self.study_event_refs = []

    def build(self, builder):
        """Build XML by appending to builder"""
        builder.start("Protocol", {})
        for child in self.study_event_refs:
            child.build(builder)
        builder.end("Protocol")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (StudyEventRef,)):
            raise ValueError('Protocol cannot accept a {0} as a child element'.format(other.__class__.__name__))
        self.set_list_attribute(other, StudyEventRef, 'study_event_refs')

        return other


class FormRef(ODMElement):
    def __init__(self, oid, order_number, mandatory):
        self.oid = oid
        self.order_number = order_number
        self.mandatory = mandatory

    def build(self, builder):
        params = dict(FormOID=self.oid,
                      OrderNumber=str(self.order_number),
                      Mandatory=bool_to_yes_no(self.mandatory)
                      )
        builder.start('FormRef', params)
        builder.end('FormRef')


class StudyEventDef(ODMElement):
    # Event types
    SCHEDULED = 'Scheduled'
    UNSCHEDULED = 'Unscheduled'
    COMMON = 'Common'

    def __init__(self, oid, name, repeating, event_type,
                 category=None,
                 access_days=None,
                 start_win_days=None,
                 target_days=None,
                 end_win_days=None,
                 overdue_days=None,
                 close_days=None
                 ):
        self.oid = oid
        self.name = name
        self.repeating = repeating
        self.event_type = event_type
        self.category = category
        self.access_days = access_days
        self.start_win_days = start_win_days
        self.target_days = target_days
        self.end_win_days = end_win_days
        self.overdue_days = overdue_days
        self.close_days = close_days
        self.formrefs = []

    def build(self, builder):
        """Build XML by appending to builder"""

        params = dict(OID=self.oid, Name=self.name,
                      Repeating=bool_to_yes_no(self.repeating),
                      Type=self.event_type)

        if self.category is not None:
            params['Category'] = self.category

        if self.access_days is not None:
            params['mdsol:AccessDays'] = str(self.access_days)

        if self.start_win_days is not None:
            params['mdsol:StartWinDays'] = str(self.start_win_days)

        if self.target_days is not None:
            params['mdsol:TargetDays'] = str(self.target_days)

        if self.end_win_days is not None:
            params['mdsol:EndWinDays'] = str(self.end_win_days)

        if self.overdue_days is not None:
            params['mdsol:OverDueDays'] = str(self.overdue_days)

        if self.close_days is not None:
            params['mdsol:CloseDays'] = str(self.close_days)

        builder.start("StudyEventDef", params)
        for formref in self.formrefs:
            formref.build(builder)
        builder.end("StudyEventDef")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (FormRef,)):
            raise ValueError('StudyEventDef cannot accept a {0} as a child element'.format(other.__class__.__name__))
        self.set_list_attribute(other, FormRef, 'formrefs')
        return other


class ItemGroupRef(ODMElement):
    def __init__(self, oid, order_number, mandatory=True):
        self.oid = oid
        self.order_number = order_number
        self.mandatory = mandatory

    def build(self, builder):
        params = dict(ItemGroupOID=self.oid,
                      OrderNumber=str(self.order_number),
                      Mandatory=bool_to_yes_no(self.mandatory),
                      )
        builder.start("ItemGroupRef", params)
        builder.end("ItemGroupRef")


class MdsolHelpText(ODMElement):
    """Help element for FormDefs and ItemDefs"""

    def __init__(self, lang, content):
        self.lang = lang
        self.content = content

    def build(self, builder):
        builder.start('mdsol:HelpText', {'xml:lang': self.lang})
        builder.data(self.content)
        builder.end('mdsol:HelpText')


class MdsolViewRestriction(ODMElement):
    """ViewRestriction for FormDefs and ItemDefs"""

    def __init__(self, rolename):
        self.rolename = rolename

    def build(self, builder):
        builder.start('mdsol:ViewRestriction', {})
        builder.data(self.rolename)
        builder.end('mdsol:ViewRestriction')


class MdsolEntryRestriction(ODMElement):
    """EntryRestriction for FormDefs and ItemDefs"""

    def __init__(self, rolename):
        self.rolename = rolename

    def build(self, builder):
        builder.start('mdsol:EntryRestriction', {})
        builder.data(self.rolename)
        builder.end('mdsol:EntryRestriction')


class FormDef(ODMElement):
    LOG_PORTRAIT = 'Portrait'
    LOG_LANDSCAPE = 'Landscape'

    DDE_MUSTNOT = 'MustNotDDE'
    DDE_MAY = 'MayDDE'
    DDE_MUST = 'MustDDE'

    NOLINK = 'NoLink'
    LINK_NEXT = 'LinkNext'
    LINK_CUSTOM = 'LinkCustom'

    def __init__(self, oid, name,
                 repeating=False,
                 order_number=None,
                 active=True,
                 template=False,
                 signature_required=False,
                 log_direction=LOG_PORTRAIT,
                 double_data_entry=DDE_MUSTNOT,
                 confirmation_style=NOLINK,
                 link_study_event_oid=None,
                 link_form_oid=None
                 ):
        self.oid = oid
        self.name = name
        self.order_number = order_number
        self.repeating = repeating  # Not actually used by Rave.
        self.active = active
        self.template = template
        self.signature_required = signature_required
        self.log_direction = log_direction
        self.double_data_entry = double_data_entry
        self.confirmation_style = confirmation_style
        self.link_study_event_oid = link_study_event_oid
        self.link_form_oid = link_form_oid
        self.itemgroup_refs = []
        self.helptexts = []  # Not clear that Rave can accept multiple from docs
        self.view_restrictions = []
        self.entry_restrictions = []

    def build(self, builder):
        params = dict(OID=self.oid,
                      Name=self.name,
                      Repeating=bool_to_yes_no(self.repeating)
                      )

        if self.order_number is not None:
            params['mdsol:OrderNumber'] = str(self.order_number)

        if self.active is not None:
            params['mdsol:Active'] = bool_to_yes_no(self.active)

        params['mdsol:Template'] = bool_to_yes_no(self.template)
        params['mdsol:SignatureRequired'] = bool_to_yes_no(self.signature_required)
        params['mdsol:LogDirection'] = self.log_direction
        params['mdsol:DoubleDataEntry'] = self.double_data_entry
        params['mdsol:ConfirmationStyle'] = self.confirmation_style

        if self.link_study_event_oid:
            params['mdsol:LinkStudyEventOID'] = self.link_study_event_oid

        if self.link_form_oid:
            params['mdsol:LinkFormOID'] = self.link_form_oid

        builder.start("FormDef", params)
        for itemgroup_ref in self.itemgroup_refs:
            itemgroup_ref.build(builder)

        for helptext in self.helptexts:
            helptext.build(builder)

        for view_restriction in self.view_restrictions:
            view_restriction.build(builder)

        for entry_restriction in self.entry_restrictions:
            entry_restriction.build(builder)
        builder.end("FormDef")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (ItemGroupRef, MdsolHelpText, MdsolViewRestriction, MdsolEntryRestriction,)):
            raise ValueError('StudyEventDef cannot accept a {0} as a child element'.format(other.__class__.__name__))

        self.set_list_attribute(other, ItemGroupRef, 'itemgroup_refs')
        self.set_list_attribute(other, MdsolHelpText, 'helptexts')
        self.set_list_attribute(other, MdsolViewRestriction, 'view_restrictions')
        self.set_list_attribute(other, MdsolEntryRestriction, 'entry_restrictions')
        return other


class MdsolLabelRef(ODMElement):
    """A reference to a label on a form"""

    def __init__(self, oid, order_number):
        self.oid = oid
        self.order_number = order_number

    def build(self, builder):
        params = dict(LabelOID=self.oid,
                      OrderNumber=str(self.order_number),
                      )

        builder.start('mdsol:LabelRef', params)
        builder.end('mdsol:LabelRef')


class MdsolAttribute(ODMElement):
    def __init__(self, namespace, name, value, transaction_type='Insert'):
        self.namespace = namespace
        self.name = name
        self.value = value
        self.transaction_type = transaction_type

    def build(self, builder):
        params = dict(Namespace=self.namespace,
                      Name=self.name,
                      Value=self.value,
                      TransactionType=self.transaction_type,
                      )

        builder.start('mdsol:Attribute', params)
        builder.end('mdsol:Attribute')


class ItemRef(ODMElement):
    def __init__(self, oid, order_number, mandatory=False, key_sequence=None,
                 imputation_method_oid=None, role=None, role_codelist_oid=None):
        self.oid = oid
        self.order_number = order_number
        self.mandatory = mandatory
        self.key_sequence = key_sequence
        self.imputation_method_oid = imputation_method_oid
        self.role = role
        self.role_codelist_oid = role_codelist_oid
        self.attributes = []

    def build(self, builder):

        params = dict(ItemOID=self.oid,
                      OrderNumber=str(self.order_number),
                      Mandatory=bool_to_yes_no(self.mandatory)
                      )

        if self.key_sequence is not None:
            params['KeySequence'] = str(self.key_sequence)

        if self.imputation_method_oid is not None:
            params['ImputationMethodOID'] = self.imputation_method_oid

        if self.role is not None:
            params['Role'] = self.role

        if self.role_codelist_oid is not None:
            params['RoleCodeListOID'] = self.role_codelist_oid

        builder.start('ItemRef', params)

        for attribute in self.attributes:
            attribute.build(builder)
        builder.end('ItemRef')

    def __lshift__(self, other):
        """ItemRef can accept MdsolAttribute(s)"""

        if not isinstance(other, (MdsolAttribute)):
            raise ValueError('ItemRef cannot accept a {0} as a child element'.format(other.__class__.__name__))
        self.set_list_attribute(other, MdsolAttribute, 'attributes')
        return other


class ItemGroupDef(ODMElement):
    def __init__(self, oid, name, repeating=False, is_reference_data=False, sas_dataset_name=None,
                 domain=None, origin=None, role=None, purpose=None, comment=None):
        self.oid = oid
        self.name = name
        self.repeating = repeating
        self.is_reference_data = is_reference_data
        self.sas_dataset_name = sas_dataset_name
        self.domain = domain
        self.origin = origin
        self.role = role
        self.purpose = purpose
        self.comment = comment
        self.item_refs = []
        self.label_refs = []

    def build(self, builder):

        params = dict(OID=self.oid,
                      Name=self.name,
                      Repeating=bool_to_yes_no(self.repeating),
                      IsReferenceData=bool_to_yes_no(self.is_reference_data)
                      )

        if self.sas_dataset_name is not None:
            params['SASDatasetName'] = self.sas_dataset_name

        if self.domain is not None:
            params['Domain'] = self.domain

        if self.origin is not None:
            params['Origin'] = self.origin

        if self.role is not None:
            params['Role'] = self.role

        if self.purpose is not None:
            params['Purpose'] = self.purpose

        if self.comment is not None:
            params['Comment'] = self.comment

        builder.start('ItemGroupDef', params)

        for itemref in self.item_refs:
            itemref.build(builder)

        # Extensions always listed AFTER core elements
        for labelref in self.label_refs:
            labelref.build(builder)
        builder.end('ItemGroupDef')

    def __lshift__(self, other):
        """ItemGroupDef can accept ItemRef and LabelRef"""

        if not isinstance(other, (ItemRef, MdsolLabelRef)):
            raise ValueError('ItemGroupDef cannot accept a {0} as a child element'.format(other.__class__.__name__))

        self.set_list_attribute(other, ItemRef, 'item_refs')
        self.set_list_attribute(other, MdsolLabelRef, 'label_refs')
        return other


class Question(ODMElement):
    def __init__(self):
        self.translations = []

    def __lshift__(self, other):
        """Override << operator"""

        if not isinstance(other, (TranslatedText)):
            raise ValueError('Question cannot accept a {0} as a child element'.format(other.__class__.__name__))
        self.set_list_attribute(other, TranslatedText, 'translations')
        return other

    def build(self, builder):
        """Questions can contain translations"""
        builder.start('Question', {})
        for translation in self.translations:
            translation.build(builder)
        builder.end('Question')


class MeasurementUnitRef(ODMElement):
    def __init__(self, oid, order_number=None):
        self.oid = oid
        self.order_number = order_number

    def build(self, builder):
        params = dict(MeasurementUnitOID=self.oid)
        if self.order_number is not None:
            params['mdsol:OrderNumber'] = str(self.order_number)

        builder.start('MeasurementUnitRef', params)
        builder.end('MeasurementUnitRef')


class MdsolHeaderText(ODMElement):
    """Header text for ItemDef when showed in grid"""

    def __init__(self, content, lang=None):
        self.content = content
        self.lang = lang

    def build(self, builder):
        params = {}
        if self.lang is not None:
            params['xml:lang'] = self.lang

        builder.start('mdsol:HeaderText', params)
        builder.data(self.content)
        builder.end('mdsol:HeaderText')


class CodeListRef(ODMElement):
    """CodeListRef: a reference a codelist within an ItemDef"""

    def __init__(self, oid):
        self.oid = oid

    def build(self, builder):
        builder.start('CodeListRef', {'CodeListOID': self.oid})
        builder.end('CodeListRef')


class MdsolLabelDef(ODMElement):
    """Label definition"""

    def __init__(self, oid, name, field_number=None):
        self.oid = oid
        self.name = name
        self.field_number = field_number
        self.help_texts = []
        self.translations = []
        self.view_restrictions = []

    def build(self, builder):
        params = dict(OID=self.oid, Name=self.name)
        if self.field_number is not None:
            params['FieldNumber'] = str(self.field_number)

        builder.start("mdsol:LabelDef", params)

        for translation in self.translations:
            translation.build(builder)

        for view_restriction in self.view_restrictions:
            view_restriction.build(builder)

        builder.end("mdsol:LabelDef")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (MdsolViewRestriction, TranslatedText)):
            raise ValueError('MdsolLabelDef cannot accept a {0} as a child element'.format(other.__class__.__name__))
        self.set_list_attribute(other, TranslatedText, 'translations')
        self.set_list_attribute(other, MdsolViewRestriction, 'view_restrictions')

        return other


class MdsolReviewGroup(ODMElement):
    """Maps to Rave review groups for an Item"""

    def __init__(self, name):
        self.name = name

    def build(self, builder):
        builder.start('mdsol:ReviewGroup', {})
        builder.data(self.name)
        builder.end('mdsol:ReviewGroup')


class CheckValue(ODMElement):
    """A value in a RangeCheck"""

    def __init__(self, value):
        self.value = value

    def build(self, builder):
        builder.start('CheckValue', {})
        builder.data(str(self.value))
        builder.end('CheckValue')


class RangeCheck(ODMElement):
    """
        Rangecheck in Rave relates to QueryHigh QueryLow and NonConformandHigh and NonComformanLow
       for other types of RangeCheck, need to use an EditCheck (part of Rave's extensions to ODM)
    """

    def __init__(self, comparator, soft_hard):
        self._comparator = None
        self.comparator = comparator
        self._soft_hard = None
        self.soft_hard = soft_hard
        self.check_value = None
        self.measurement_unit_ref = None

    @property
    def comparator(self):
        return self._comparator

    @comparator.setter
    def comparator(self, value):
        if not isinstance(value, RangeCheckComparatorType):
            raise AttributeError("%s comparator is invalid in RangeCheck." % (value,))
        self._comparator = value

    @property
    def soft_hard(self):
        return self._soft_hard

    @soft_hard.setter
    def soft_hard(self, value):
        if not isinstance(value, RangeCheckType):
            raise AttributeError("%s soft_hard invalid in RangeCheck." % (value,))
        self._soft_hard = value

    def build(self, builder):
        params = dict(SoftHard=self.soft_hard.value, Comparator=self.comparator.value)
        builder.start("RangeCheck", params)
        if self.check_value is not None:
            self.check_value.build(builder)
        if self.measurement_unit_ref is not None:
            self.measurement_unit_ref.build(builder)
        builder.end("RangeCheck")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (CheckValue, MeasurementUnitRef,)):
            raise ValueError('RangeCheck cannot accept a {0} as a child element'.format(other.__class__.__name__))

        self.set_single_attribute(other, CheckValue, 'check_value')
        self.set_single_attribute(other, MeasurementUnitRef, 'measurement_unit_ref')


class ItemDef(ODMElement):
    VALID_DATATYPES = [DataType.Text, DataType.Integer, DataType.Float, DataType.Date,
                       DataType.DateTime, DataType.Time]

    def __init__(self, oid, name, datatype, length,
                 significant_digits=None,
                 sas_field_name=None,
                 sds_var_name=None,
                 origin=None,  # Not mapped in Rave
                 comment=None,
                 active=True,
                 control_type=None,
                 acceptable_file_extensions=None,
                 indent_level=0,
                 source_document_verify=False,
                 default_value=None,
                 sas_format=None,
                 sas_label=None,
                 query_future_date=False,
                 visible=True,
                 translation_required=False,
                 query_non_conformance=False,
                 other_visits=False,
                 can_set_item_group_date=False,
                 can_set_form_date=False,
                 can_set_study_event_date=False,
                 can_set_subject_date=False,
                 visual_verify=False,
                 does_not_break_signature=False,
                 date_time_format=None,
                 field_number=None,
                 variable_oid=None
                 ):
        self.oid = oid
        self.name = name

        if datatype not in ItemDef.VALID_DATATYPES:
            raise AttributeError('{0} is not a valid datatype!'.format(datatype))

        if control_type is not None:
            if not isinstance(control_type, ControlType):
                raise AttributeError("{0} is not a valid Control Type".format(control_type))

        self.datatype = datatype
        self.length = length
        self.significant_digits = significant_digits
        self.sas_field_name = sas_field_name
        self.sds_var_name = sds_var_name
        self.origin = origin
        self.comment = comment
        self.active = active
        self.control_type = control_type
        self.acceptable_file_extensions = acceptable_file_extensions
        self.indent_level = indent_level
        self.source_document_verify = source_document_verify
        self.default_value = default_value
        self.sas_format = sas_format
        self.sas_label = sas_label
        self.query_future_date = query_future_date
        self.visible = visible
        self.translation_required = translation_required
        self.query_non_conformance = query_non_conformance
        self.other_visits = other_visits
        self.can_set_item_group_date = can_set_item_group_date
        self.can_set_form_date = can_set_form_date
        self.can_set_study_event_date = can_set_study_event_date
        self.can_set_subject_date = can_set_subject_date
        self.visual_verify = visual_verify
        self.does_not_break_signature = does_not_break_signature
        self.date_time_format = date_time_format
        self.field_number = field_number
        self.variable_oid = variable_oid

        self.question = None
        self.codelistref = None
        self.measurement_unit_refs = []
        self.help_texts = []
        self.view_restrictions = []
        self.entry_restrictions = []
        self.header_text = None
        self.review_groups = []
        self.range_checks = []

    def build(self, builder):
        """Build XML by appending to builder"""

        params = dict(OID=self.oid,
                      Name=self.name,
                      DataType=self.datatype.value,
                      Length=str(self.length),
                      )

        if self.date_time_format is not None:
            params['mdsol:DateTimeFormat'] = self.date_time_format

        params['mdsol:Active'] = bool_to_yes_no(self.active)

        if self.significant_digits is not None:
            params['SignificantDigits'] = str(self.significant_digits)

        if self.sas_field_name is not None:
            params['SASFieldName'] = self.sas_field_name

        if self.sds_var_name is not None:
            params['SDSVarName'] = self.sds_var_name

        if self.origin is not None:
            params['Origin'] = self.origin

        if self.comment is not None:
            params['Comment'] = self.comment

        if self.control_type is not None:
            params['mdsol:ControlType'] = self.control_type.value

        if self.acceptable_file_extensions is not None:
            params['mdsol:AcceptableFileExtensions'] = self.acceptable_file_extensions

        if self.default_value is not None:
            params['mdsol:DefaultValue'] = str(self.default_value)

        params['mdsol:SourceDocument'] = bool_to_yes_no(self.source_document_verify)
        params['mdsol:IndentLevel'] = str(self.indent_level)

        if self.sas_format is not None:
            params['mdsol:SASFormat'] = self.sas_format

        if self.sas_label is not None:
            params['mdsol:SASLabel'] = self.sas_label

        params['mdsol:QueryFutureDate'] = bool_to_yes_no(self.query_future_date)
        params['mdsol:Visible'] = bool_to_yes_no(self.visible)
        params['mdsol:TranslationRequired'] = bool_to_yes_no(self.translation_required)
        params['mdsol:QueryNonConformance'] = bool_to_yes_no(self.query_non_conformance)
        params['mdsol:OtherVisits'] = bool_to_yes_no(self.other_visits)
        params['mdsol:CanSetItemGroupDate'] = bool_to_yes_no(self.can_set_item_group_date)
        params['mdsol:CanSetFormDate'] = bool_to_yes_no(self.can_set_form_date)
        params['mdsol:CanSetStudyEventDate'] = bool_to_yes_no(self.can_set_study_event_date)
        params['mdsol:CanSetSubjectDate'] = bool_to_yes_no(self.can_set_subject_date)
        params['mdsol:VisualVerify'] = bool_to_yes_no(self.visual_verify)
        params['mdsol:DoesNotBreakSignature'] = bool_to_yes_no(self.does_not_break_signature)

        if self.field_number is not None:
            params['mdsol:FieldNumber'] = self.field_number

        if self.variable_oid is not None:
            params['mdsol:VariableOID'] = self.variable_oid

        builder.start("ItemDef", params)

        if self.question is not None:
            self.question.build(builder)

        if self.codelistref is not None:
            self.codelistref.build(builder)

        for mur in self.measurement_unit_refs:
            mur.build(builder)

        for range_check in self.range_checks:
            range_check.build(builder)

        if self.header_text is not None:
            self.header_text.build(builder)

        for view_restriction in self.view_restrictions:
            view_restriction.build(builder)

        for entry_restriction in self.entry_restrictions:
            entry_restriction.build(builder)

        for help_text in self.help_texts:
            help_text.build(builder)

        for review_group in self.review_groups:
            review_group.build(builder)

        builder.end("ItemDef")

    def __lshift__(self, other):
        """Override << operator"""

        # ExternalQuestion?,,
        # Role*, Alias*,
        # mdsol:HelpText?, mdsol:ViewRestriction* or mdsolEntryRestrictions*), (or mdsol:ReviewGroups*), mdsol:Label?)

        if not isinstance(other, (MdsolHelpText, MdsolEntryRestriction, MdsolViewRestriction, Question,
                                  MeasurementUnitRef, CodeListRef, MdsolHeaderText, MdsolReviewGroup, RangeCheck)):
            raise ValueError('ItemDef cannot accept a {0} as a child element'.format(other.__class__.__name__))

        self.set_single_attribute(other, Question, 'question')
        self.set_single_attribute(other, CodeListRef, 'codelistref')
        self.set_single_attribute(other, MdsolHeaderText, 'header_text')
        self.set_list_attribute(other, RangeCheck, 'range_checks')
        self.set_list_attribute(other, MeasurementUnitRef, 'measurement_unit_refs')
        self.set_list_attribute(other, MdsolHelpText, 'help_texts')
        self.set_list_attribute(other, MdsolViewRestriction, 'view_restrictions')
        self.set_list_attribute(other, MdsolEntryRestriction, 'entry_restrictions')
        self.set_list_attribute(other, MdsolReviewGroup, 'review_groups')
        return other


class Decode(ODMElement):
    def __init__(self):
        self.translations = []

    def build(self, builder):
        builder.start("Decode", {})
        for translation in self.translations:
            translation.build(builder)
        builder.end("Decode")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, TranslatedText):
            raise ValueError('Decode cannot accept child of type {0}'.format(other.__class__.__name__))
        self.translations.append(other)
        return other


class CodeListItem(ODMElement):
    def __init__(self, coded_value, order_number=None, specify=False):
        self.coded_value = coded_value
        self.order_number = order_number
        self.specify = specify
        self.decode = None

    def build(self, builder):
        params = dict(CodedValue=self.coded_value)
        if self.order_number is not None:
            params['mdsol:OrderNumber'] = str(self.order_number)

        if self.specify:
            params['mdsol:Specify'] = "Yes"

        builder.start("CodeListItem", params)
        if self.decode is not None:
            self.decode.build(builder)
        builder.end("CodeListItem")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, Decode):
            raise ValueError('CodelistItem cannot accept child of type {0}'.format(other.__class__.__name__))
        self.set_single_attribute(other, Decode, 'decode')
        return other


class CodeList(ODMElement):
    """A container for CodeListItems equivalent of Rave Dictionary"""
    VALID_DATATYPES = [DataType.Integer, DataType.Text, DataType.Float, DataType.String]

    def __init__(self, oid, name, datatype, sas_format_name=None):
        self.oid = oid
        self.name = name
        if datatype not in CodeList.VALID_DATATYPES:
            raise ValueError("{0} is not a valid CodeList datatype".format(datatype))
        self.datatype = datatype
        self.sas_format_name = sas_format_name
        self.codelist_items = []

    def build(self, builder):
        params = dict(OID=self.oid,
                      Name=self.name,
                      DataType=self.datatype.value)
        if self.sas_format_name is not None:
            params['SASFormatName'] = self.sas_format_name
        builder.start("CodeList", params)
        for item in self.codelist_items:
            item.build(builder)
        builder.end("CodeList")

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, CodeListItem):
            raise ValueError('Codelist cannot accept child of type {0}'.format(other.__class__.__name__))
        self.set_list_attribute(other, CodeListItem, 'codelist_items')

        return other


class MdsolConfirmationMessage(ODMElement):
    """Form is saved confirmation message"""

    def __init__(self, message, lang=None):
        self.message = message
        self.lang = lang

    def build(self, builder):
        params = {}
        if self.lang:
            params['xml:lang'] = self.lang
        builder.start('mdsol:ConfirmationMessage', params)
        builder.data(self.message)
        builder.end('mdsol:ConfirmationMessage')


class MdsolDerivationStep(ODMElement):
    """A derivation step modeled after the Architect Loader definition.
       Do not use directly, use appropriate subclasses.
    """
    VALID_STEPS = VALID_DERIVATION_STEPS

    def __init__(self,
                 variable_oid=None,
                 data_format=None,
                 form_oid=None,
                 folder_oid=None,
                 field_oid=None,
                 value=None,
                 function=None,
                 custom_function=None,
                 record_position=None,
                 form_repeat_number=None,
                 folder_repeat_number=None,
                 logical_record_position=None
                 ):

        self.variable_oid = variable_oid
        self.data_format = data_format
        self.form_oid = form_oid
        self.folder_oid = folder_oid
        self.field_oid = field_oid
        self.value = value
        self._function = None
        self.function = function
        self.custom_function = custom_function
        self.record_position = record_position
        self.form_repeat_number = form_repeat_number
        self.folder_repeat_number = folder_repeat_number
        self.logical_record_position = logical_record_position

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, value):
        if value is not None:
            if value not in MdsolDerivationStep.VALID_STEPS:
                raise AttributeError("Invalid derivation function %s" % value)
        self._function = value

    def build(self, builder):
        params = dict()

        if self.variable_oid is not None:
            params['VariableOID'] = self.variable_oid

        if self.data_format is not None:
            params['DataFormat'] = self.data_format

        if self.folder_oid is not None:
            params['FolderOID'] = self.folder_oid

        if self.field_oid is not None:
            params['FieldOID'] = self.field_oid

        if self.form_oid is not None:
            params['FormOID'] = self.form_oid

        if self.value is not None:
            params['Value'] = self.value

        if self.function is not None:
            params['Function'] = self.function.value

        if self.custom_function is not None:
            params['CustomFunction'] = self.custom_function

        if self.record_position is not None:
            params['RecordPosition'] = str(self.record_position)

        if self.form_repeat_number is not None:
            params['FormRepeatNumber'] = str(self.form_repeat_number)

        if self.folder_repeat_number is not None:
            params['FolderRepeatNumber'] = str(self.folder_repeat_number)

        if self.logical_record_position is not None:
            params['LogicalRecordPosition'] = self.logical_record_position

        builder.start("mdsol:DerivationStep", params)
        builder.end("mdsol:DerivationStep")


class MdsolCheckStep(ODMElement):
    """A check step modeled after the Architect Loader definition.
       Do not use directly, use appropriate subclasses.
    """
    VALID_STEPS = ALL_STEPS

    def __init__(self,
                 variable_oid=None,
                 data_format=None,
                 form_oid=None,
                 folder_oid=None,
                 field_oid=None,
                 static_value=None,
                 function=None,
                 custom_function=None,
                 record_position=None,
                 form_repeat_number=None,
                 folder_repeat_number=None,
                 logical_record_position=None
                 ):

        self.variable_oid = variable_oid
        self.data_format = data_format
        self.form_oid = form_oid
        self.folder_oid = folder_oid
        self.field_oid = field_oid
        self.static_value = static_value
        self._function = None
        self.function = function
        self.custom_function = custom_function
        self.record_position = record_position
        self.form_repeat_number = form_repeat_number
        self.folder_repeat_number = folder_repeat_number
        self.logical_record_position = logical_record_position

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, value):
        if value is not None:
            if value not in MdsolCheckStep.VALID_STEPS:
                raise AttributeError("Invalid function %s" % value)
        self._function = value

    def build(self, builder):
        params = dict()

        if self.variable_oid is not None:
            params['VariableOID'] = self.variable_oid

        if self.data_format is not None:
            params['DataFormat'] = self.data_format

        if self.folder_oid is not None:
            params['FolderOID'] = self.folder_oid

        if self.field_oid is not None:
            params['FieldOID'] = self.field_oid

        if self.form_oid is not None:
            params['FormOID'] = self.form_oid

        if self.static_value is not None:
            params['StaticValue'] = self.static_value

        if self.function is not None:
            params['Function'] = self.function.value

        if self.custom_function is not None:
            params['CustomFunction'] = self.custom_function

        if self.record_position is not None:
            params['RecordPosition'] = str(self.record_position)

        if self.form_repeat_number is not None:
            params['FormRepeatNumber'] = str(self.form_repeat_number)

        if self.folder_repeat_number is not None:
            params['FolderRepeatNumber'] = str(self.folder_repeat_number)

        if self.logical_record_position is not None:
            params['LogicalRecordPosition'] = self.logical_record_position

        builder.start("mdsol:CheckStep", params)
        builder.end("mdsol:CheckStep")


class MdsolCheckAction(ODMElement):
    """
    Check Action modeled after check action in Architect Loader spreadsheet.
    Do not use directly, use appropriate sub-class.
    """

    def __init__(self,
                 variable_oid=None,
                 field_oid=None,
                 form_oid=None,
                 folder_oid=None,
                 record_position=None,
                 form_repeat_number=None,
                 folder_repeat_number=None,
                 check_action_type=None,
                 check_string=None,
                 check_options=None,
                 check_script=None
                 ):

        self.variable_oid = variable_oid
        self.folder_oid = folder_oid
        self.field_oid = field_oid
        self.form_oid = form_oid
        self.record_position = record_position
        self.form_repeat_number = form_repeat_number
        self.folder_repeat_number = folder_repeat_number
        self._check_action_type = None
        self.check_action_type = check_action_type
        self.check_string = check_string
        self.check_options = check_options
        self.check_script = check_script

    @property
    def check_action_type(self):
        return self._check_action_type

    @check_action_type.setter
    def check_action_type(self, value):
        if value is not None:
            if not isinstance(value, ActionType):
                raise AttributeError("Invalid check action %s" % value)
        self._check_action_type = value

    def build(self, builder):
        params = dict()

        if self.variable_oid is not None:
            params['VariableOID'] = self.variable_oid

        if self.field_oid is not None:
            params['FieldOID'] = self.field_oid

        if self.form_oid is not None:
            params['FormOID'] = self.form_oid

        if self.folder_oid is not None:
            params['FolderOID'] = self.folder_oid

        if self.record_position is not None:
            params['RecordPosition'] = str(self.record_position)

        if self.form_repeat_number is not None:
            params['FormRepeatNumber'] = str(self.form_repeat_number)

        if self.folder_repeat_number is not None:
            params['FolderRepeatNumber'] = str(self.folder_repeat_number)

        if self.check_action_type is not None:
            params['Type'] = self.check_action_type.value

        if self.check_string is not None:
            params['String'] = self.check_string

        if self.check_options is not None:
            params['Options'] = self.check_options

        if self.check_script is not None:
            params['Script'] = self.check_script

        builder.start("mdsol:CheckAction", params)
        builder.end("mdsol:CheckAction")


class MdsolEditCheckDef(ODMElement):
    """Extension for Rave edit checks"""

    def __init__(self, oid, active=True, bypass_during_migration=False, needs_retesting=False):
        self.oid = oid
        self.active = active
        self.bypass_during_migration = bypass_during_migration
        self.needs_retesting = needs_retesting
        self.check_steps = []
        self.check_actions = []

    def build(self, builder):
        params = dict(OID=self.oid,
                      Active=bool_to_true_false(self.active),
                      BypassDuringMigration=bool_to_true_false(self.bypass_during_migration),
                      NeedsRetesting=bool_to_true_false(self.needs_retesting)
                      )

        builder.start('mdsol:EditCheckDef', params)
        for step in self.check_steps:
            step.build(builder)

        for action in self.check_actions:
            action.build(builder)
        builder.end('mdsol:EditCheckDef')

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (MdsolCheckStep, MdsolCheckAction,)):
            raise ValueError('EditCheck cannot accept a {0} as a child element'.format(other.__class__.__name__))
        self.set_list_attribute(other, MdsolCheckStep, 'check_steps')
        self.set_list_attribute(other, MdsolCheckAction, 'check_actions')


class MdsolDerivationDef(ODMElement):
    """Extension for Rave derivations"""

    def __init__(self, oid, active=True,
                 bypass_during_migration=False,
                 needs_retesting=False,
                 variable_oid=None,
                 field_oid=None,
                 form_oid=None,
                 folder_oid=None,
                 record_position=None,
                 form_repeat_number=None,
                 folder_repeat_number=None,
                 logical_record_position=None,
                 all_variables_in_folders=None,
                 all_variables_in_fields=None
                 ):
        self.oid = oid
        self.active = active
        self.bypass_during_migration = bypass_during_migration
        self.needs_retesting = needs_retesting
        self.variable_oid = variable_oid
        self.field_oid = field_oid
        self.form_oid = form_oid
        self.folder_oid = folder_oid
        self.record_position = record_position
        self.form_repeat_number = form_repeat_number
        self.folder_repeat_number = folder_repeat_number
        self.logical_record_position = logical_record_position
        self.all_variables_in_folders = all_variables_in_folders
        self.all_variables_in_fields = all_variables_in_fields
        self.derivation_steps = []

    def build(self, builder):
        params = dict(
            OID=self.oid,
            Active=bool_to_true_false(self.active),
            BypassDuringMigration=bool_to_true_false(self.bypass_during_migration),
            NeedsRetesting=bool_to_true_false(self.needs_retesting)
        )

        if self.variable_oid is not None:
            params['VariableOID'] = self.variable_oid

        if self.field_oid is not None:
            params['FieldOID'] = self.field_oid

        if self.form_oid is not None:
            params['FormOID'] = self.form_oid

        if self.folder_oid is not None:
            params['FolderOID'] = self.folder_oid

        if self.record_position is not None:
            params['RecordPosition'] = str(self.record_position)

        if self.form_repeat_number is not None:
            params['FormRepeatNumber'] = str(self.form_repeat_number)

        if self.folder_repeat_number is not None:
            params['FolderRepeatNumber'] = str(self.folder_repeat_number)

        if self.all_variables_in_folders is not None:
            params['AllVariablesInFolders'] = bool_to_true_false(self.all_variables_in_folders)

        if self.all_variables_in_fields is not None:
            params['AllVariablesInFields'] = bool_to_true_false(self.all_variables_in_fields)

        if self.logical_record_position is not None:
            params['LogicalRecordPosition'] = self.logical_record_position

        builder.start('mdsol:DerivationDef', params)
        for step in self.derivation_steps:
            step.build(builder)
        builder.end('mdsol:DerivationDef')

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, MdsolDerivationStep):
            raise ValueError('Derivation cannot accept a {0} as a child element'.format(other.__class__.__name__))
        self.set_list_attribute(other, MdsolDerivationStep, 'derivation_steps')


class MdsolCustomFunctionDef(ODMElement):
    """Extension for Rave Custom functions"""
    VB = "VB"  # VB was deprecated in later Rave versions.
    C_SHARP = "C#"
    SQL = "SQ"
    VALID_LANGUAGES = [C_SHARP, SQL, VB]

    def __init__(self, oid, code, language="C#"):
        self.oid = oid
        self.code = code
        self.language = language

    def build(self, builder):
        params = dict(OID=self.oid, Language=self.language)
        builder.start('mdsol:CustomFunctionDef', params)
        builder.data(self.code)
        builder.end('mdsol:CustomFunctionDef')


class MetaDataVersion(ODMElement):
    """MetaDataVersion, child of study"""

    def __init__(self, oid, name,
                 description=None,
                 primary_formoid=None,
                 default_matrix_oid=None,
                 delete_existing=False,
                 signature_prompt=None):
        self.oid = oid
        self.name = name
        self.description = description
        self.primary_formoid = primary_formoid
        self.default_matrix_oid = default_matrix_oid
        self.delete_existing = delete_existing
        self.signature_prompt = signature_prompt
        self.confirmation_message = None
        self.protocol = None
        self.codelists = []
        self.item_defs = []
        self.label_defs = []
        self.item_group_defs = []
        self.form_defs = []
        self.study_event_defs = []
        self.edit_checks = []
        self.derivations = []
        self.custom_functions = []

    def build(self, builder):
        """Build XML by appending to builder"""

        params = dict(OID=self.oid, Name=self.name)

        if self.description is not None:
            params['Description'] = self.description

        if self.signature_prompt is not None:
            params['mdsol:SignaturePrompt'] = self.signature_prompt

        if self.primary_formoid is not None:
            params['mdsol:PrimaryFormOID'] = self.primary_formoid

        if self.default_matrix_oid is not None:
            params['mdsol:DefaultMatrixOID'] = self.default_matrix_oid

        params['mdsol:DeleteExisting'] = bool_to_yes_no(self.delete_existing)

        builder.start("MetaDataVersion", params)
        if self.protocol:
            self.protocol.build(builder)

        for event in self.study_event_defs:
            event.build(builder)

        for formdef in self.form_defs:
            formdef.build(builder)

        for itemgroupdef in self.item_group_defs:
            itemgroupdef.build(builder)

        for itemdef in self.item_defs:
            itemdef.build(builder)

        for codelist in self.codelists:
            codelist.build(builder)

        # Extensions must always come after core elements
        if self.confirmation_message:
            self.confirmation_message.build(builder)

        for labeldef in self.label_defs:
            labeldef.build(builder)

        for edit_check in self.edit_checks:
            edit_check.build(builder)

        for derivation in self.derivations:
            derivation.build(builder)

        for custom_function in self.custom_functions:
            custom_function.build(builder)

        builder.end("MetaDataVersion")

    def __lshift__(self, other):
        """Override << operator"""

        if not isinstance(other, (Protocol, StudyEventDef, FormDef, ItemGroupDef, ItemDef, MdsolLabelDef, CodeList,
                                  MdsolConfirmationMessage, MdsolEditCheckDef, MdsolDerivationDef,
                                  MdsolCustomFunctionDef)):
            raise ValueError('MetaDataVersion cannot accept a {0} as a child element'.format(other.__class__.__name__))

        self.set_single_attribute(other, Protocol, 'protocol')
        self.set_single_attribute(other, MdsolConfirmationMessage, 'confirmation_message')
        self.set_list_attribute(other, StudyEventDef, 'study_event_defs')
        self.set_list_attribute(other, FormDef, 'form_defs')
        self.set_list_attribute(other, ItemGroupDef, 'item_group_defs')
        self.set_list_attribute(other, MdsolLabelDef, 'label_defs')
        self.set_list_attribute(other, ItemDef, 'item_defs')
        self.set_list_attribute(other, CodeList, 'codelists')
        self.set_list_attribute(other, MdsolEditCheckDef, 'edit_checks')
        self.set_list_attribute(other, MdsolDerivationDef, 'derivations')
        self.set_list_attribute(other, MdsolCustomFunctionDef, 'custom_functions')  # NB. Current schema limits to 1
        return other


class Study(ODMElement):
    """ODM Study Metadata element"""

    PROJECT = 'Project'
    GLOBAL_LIBRARY = 'GlobalLibrary Volume'
    PROJECT_TYPES = [PROJECT, GLOBAL_LIBRARY]

    def __init__(self, oid, project_type=None):
        self.oid = oid
        self.global_variables = None
        self.basic_definitions = None
        self.metadata_version = None
        self.studyevent_defs = []
        if project_type is None:
            self.project_type = "Project"
        else:
            if project_type in Study.PROJECT_TYPES:
                self.project_type = project_type
            else:
                raise ValueError('Project type "{0}" not valid. Expected one of {1}'.format(project_type,
                                                                                            ','.join(
                                                                                                Study.PROJECT_TYPES)))

    def __lshift__(self, other):
        """Override << operator"""

        if not isinstance(other, (GlobalVariables, BasicDefinitions, MetaDataVersion)):
            raise ValueError('Study cannot accept a {0} as a child element'.format(other.__class__.__name__))

        self.set_single_attribute(other, GlobalVariables, 'global_variables')
        self.set_single_attribute(other, BasicDefinitions, 'basic_definitions')
        self.set_single_attribute(other, MetaDataVersion, 'metadata_version')

        return other

    def build(self, builder):
        """Build XML by appending to builder"""
        params = dict(OID=self.oid)
        params['mdsol:ProjectType'] = self.project_type

        builder.start("Study", params)

        # Ask children
        if self.global_variables is not None:
            self.global_variables.build(builder)

        if self.basic_definitions is not None:
            self.basic_definitions.build(builder)

        if self.metadata_version is not None:
            self.metadata_version.build(builder)

        builder.end("Study")



# -*- coding: utf-8 -*-
__author__ = 'isparks'

from rwslib.builders import *

def example_clinical_data(study_name, environment):
    """Test demonstrating building clinical data"""

    odm = ODM("test system")(
       ClinicalData("Mediflex", "DEV")(
          SubjectData("MDSOL", "IJS TEST4", transaction_type="Insert")(
             StudyEventData("SUBJECT")(
                FormData("EN", transaction_type="Update")(
                   ItemGroupData()(
                      ItemData("SUBJINIT", "AAA")(
                            AuditRecord(edit_point=AuditRecord.EDIT_DATA_MANAGEMENT,
                                      used_imputation_method= False,
                                      identifier='X2011',
                                      include_file_oid=False)(
                                            UserRef("isparks"),
                                            LocationRef("MDSOL"),
                                            ReasonForChange("Data Entry Error"),
                                            DateTimeStamp(datetime(2015, 9, 11, 10, 15, 22, 80))
                            ),
                            MdsolQuery(value="Subject intials should be 2 chars only.", recipient="Site from System",
                                       status=QueryStatusType.Open)
                      ),
                      ItemData("SUBJID", '001')
                   )
                )
             )
          )
       )
    )
    return odm


def example_metadata(study_name, draft_name):
    """Example of building a metadata doc"""
    odm = ODM("SYSTEM_NAME", filetype=ODM.FILETYPE_SNAPSHOT)

    study = Study(study_name, project_type=Study.PROJECT)

    # Push study element into odm
    odm << study

    # Create global variables and set them into study.
    study << GlobalVariables(study_name) #Expected that protocol name will match the Study OID.

    # Create some basic definitions
    bd = BasicDefinitions()

    # Add some measurement units to the basic definitions. This time using the call () syntax:
    bd(
        MeasurementUnit("KG", "Kilograms")(
            Symbol()(TranslatedText("Kilograms"))
        ),
        MeasurementUnit("CM", "Centimeters")(
            Symbol()(TranslatedText("Centimeters"))
        )
    )

    # Add basic definitions to study
    study << bd

    # Now metadata which will contain all our form and field defs eventually
    meta = MetaDataVersion('META1', draft_name)
    study << meta

    # Protocol contains StudyEventRefs
    protocol = Protocol()
    # Add some StudyEventRefs
    protocol << StudyEventRef("FLDR1", 1, True)  # Order 1, Mandatory
    # protocol << StudyEventRef("FLDR2", 2, False) # Order 2, Not Mandatory
    # protocol << StudyEventRef("AE", 3, True)

    meta << protocol

    # Add Study Event Defs with some child FormRefs
    fldr1 = StudyEventDef("FLDR1", "Folder 1", False, StudyEventDef.SCHEDULED)

    fldr1 << FormRef("DM", 1, True)
    fldr1 << FormRef("VS", 2, True)

    meta << fldr1

    meta << StudyEventDef("FLDR2", "Folder 2", False, StudyEventDef.UNSCHEDULED)(
        FormRef("VS", 1, True)
    )

    meta << StudyEventDef("AE", "Adverse Events", False, StudyEventDef.COMMON)(
        FormRef("AE", 1, False)
    )

    dm_form = FormDef("DM","Demography")
    dm_form << MdsolHelpText("en","Some help text for Demography")
    dm_form << MdsolViewRestriction('Data Manager')
    dm_form << MdsolEntryRestriction('Batch Upload')
    dm_form << ItemGroupRef("DM_IG1", 1)
    dm_form << ItemGroupRef("DM_IG2", 2)

    # Add to metadata
    meta << dm_form

    # Define item group
    meta << ItemGroupDef("DM_IG1", "DM Item Group 1")(
        MdsolLabelRef("LBL1", 1),
        ItemRef("SEX", 2)(
            MdsolAttribute("Standards","CDASH","SEX"),
            MdsolAttribute("Standards","STDNUMBER","1120")
        ),
        ItemRef("RACE", 3),
        ItemRef("RACE_OTH", 4),
        ItemRef("DOB", 5),
        ItemRef("AGE", 6)
    )

    # Add the ItemDefs
    meta << ItemDef("SEX", "Gender", DataType.Text, 1, control_type=ControlType.RadioButton
       )(
        Question()(TranslatedText("Gender at Birth")),
        CodeListRef("CL_SEX")
    )
    meta << ItemDef("RACE", "Race", DataType.Text, 2,
                    control_type=ControlType.RadioButtonVertical
                    )(
        Question()(TranslatedText("Race")),
        CodeListRef("CL_RACE")
    )
    meta << ItemDef("RACE_OTH", "RaceOther", DataType.Text, 20) \
           << Question() << TranslatedText("If Race Other, please specify")

    meta << ItemDef("DOB", "DateOfBirth", DataType.Date, 10,
                    control_type=ControlType.DateTime,
                    date_time_format="dd/mm/yyyy"
                    )(
        Question()(TranslatedText("Date of Birth")),
        MdsolHelpText("en","If month unknown, enter January")
    )

    meta << ItemDef("AGE", "Age in Years", DataType.Integer, 4, significant_digits=3, control_type=ControlType.Text
       )(
        Question()(TranslatedText("Age in Years")),
        RangeCheck(RangeCheckComparatorType.GreaterThanEqualTo, RangeCheckType.Soft) (
            CheckValue("18")
        ),
        RangeCheck(RangeCheckComparatorType.LessThanEqualTo, RangeCheckType.Soft) (
            CheckValue("65")
        )
    )

    # Add a Label
    meta.add(MdsolLabelDef("LBL1", "Label1")(TranslatedText("Please answer all questions.")))

    # As well as () and << you can use add()
    meta.add(
        CodeList("CL_SEX", "SEX", datatype=DataType.Text)(
            CodeListItem("M").add(
                Decode().add(
                    TranslatedText("Male"))
            ),
            CodeListItem("F").add(
                Decode().add(
                    TranslatedText("Female"))
            ),
        ),
        CodeList("CL_RACE", "RACE", datatype=DataType.Text)(
            CodeListItem("AS")(Decode()(TranslatedText("Asian"))),
            CodeListItem("CA")(Decode()(TranslatedText("White"))),
            CodeListItem("OT")(Decode()(TranslatedText("Other"))),
        )
    )

    meta.add(MdsolEditCheckDef('CHECK1')(
                # Static value required to make this stick, gets ignored but won't load without it
                MdsolCheckStep(form_oid="DM", field_oid="RACE", data_format='CodedValue', static_value="1"),
                MdsolCheckStep(static_value="OT", data_format="$2"),
                MdsolCheckStep(function=StepType.IsEqualTo),
                MdsolCheckStep(form_oid="DM", field_oid="RACE_OTH"),
                MdsolCheckStep(function=StepType.IsEmpty),
                MdsolCheckStep(function=StepType.And),

                MdsolCheckAction(form_oid="DM", field_oid="RACE_OTH",
                                 check_action_type=ActionType.OpenQuery,
                                 check_string="Race is set as OTHER but not specified. Please correct.",
                                 check_options="Site from System,RequiresResponse,RequiresManualClose"
                                 )

             ),
             MdsolEditCheckDef('CHECK2')
    )

    meta.add(MdsolCustomFunctionDef("CF1","SELECT 1,2 FROM DataPoints", language="SQ"))
    meta.add(MdsolCustomFunctionDef("CF2","return true;", language="C#"))

    meta.add(
            # Variable OID required
            MdsolDerivationDef("AGE", form_oid="DM", field_oid="AGE")(
            # Variable OID required to be identified as a data step
            MdsolDerivationStep(form_oid="DM", field_oid="DOB", data_format="StandardValue", variable_oid="DOB"),
            MdsolDerivationStep(function=StepType.Age)
        )
    )

    return odm

if __name__ == '__main__':
    from rwslib import RWSConnection
    from rwslib.rws_requests import PostMetadataRequest, PostDataRequest
    from _settings import accounts
    account = accounts['innovate']
    r = RWSConnection('innovate', account['username'], account['password'])


    if False: # MetaData
        projectname = 'TESTSTUDY'
        odm_definition = example_metadata(projectname, "Draft1")
        request = PostMetadataRequest(projectname, str(odm_definition))
    else: #Clinical Data
        projectname = 'Mediflex'
        odm_definition = example_clinical_data(projectname,"DEV")
        request = PostDataRequest(str(odm_definition))
    print str(odm_definition)

    response = r.send_request(request)
    print(str(response))


__author__ = 'andrew'

from rwslib import RWSConnection
from rwslib.rws_requests import *
from rwslib.rws_requests.odm_adapter import *
from rwslib.rws_requests.biostats_gateway import *

if __name__ == '__main__':

    from _settings import accounts

    acc = accounts['innovate']
    rave = RWSConnection('innovate', acc['username'], acc['password'])

    print(rave.send_request(VersionRequest()))

    audits = rave.send_request(AuditRecordsRequest('Mediflex','Dev',startid=4000000,per_page=10000))
    print rave.next_link # Get headers, next and last entries?
    #print audits
    while rave.next_link <> None:
        rave.next(AuditRecordsRequest('Mediflex','Dev'))
        print(rave.next_link)


# -*- coding: utf-8 -*-
__author__ = 'isparks'

accounts = {'innovate' : {'username': 'isparks',
                         'password':'Jupiter103'},
            'otherurl' : {'username': 'username',
                         'password':'password'},
           }


__author__ = 'isparks'

from lxml import etree

MEDI_NS = '{http://www.mdsol.com/ns/odm/metadata}'
ODM_NS = '{http://www.cdisc.org/ns/odm/v1.3}'


def getEnvironmentFromNameAndProtocol(studyname, protocolname):
    """Extract environment name using studyname and protocolname to guide"""
    # StudyName =    "TEST (1) (DEV)"
    # ProtocolName = "TEST (1)"
    # Raw Env      =           "(DEV)"

    raw_env = studyname[len(protocolname):].strip()

    if '(' in raw_env:
        l_brace_pos = raw_env.rfind('(')
        r_brace_pos = raw_env.rfind(')')
        return raw_env[l_brace_pos + 1:r_brace_pos]
    else:
        return raw_env


def parseXMLString(xml):
    """Parse XML string, return root"""

    # Remove BOM if it exists (different requests seem to have different BOMs)
    unichr_captured = ''

    if not xml.strip():
        return u''

    while xml[0] != u'<':
        unichr_captured += xml[0]
        xml = xml[1:]
    parser = etree.XMLParser(ns_clean=True, collect_ids=False, huge_tree=True)
    try:
        return etree.fromstring(xml.encode('utf-8'), parser=parser)
    except etree.XMLSyntaxError:
        raise Exception(xml)


class RWSException(Exception):
    """RWS Exception. Usual to attach the error response object"""

    def __init__(self, msg, rws_error):
        Exception.__init__(self, msg)
        self.rws_error = rws_error


class XMLRepr(object):
    """Classes that represent objects passed back from RWS as XML"""

    def __init__(self, xml):
        self.root = parseXMLString(xml)

    def __str__(self):
        """String representation of same"""
        return etree.tostring(self.root, encoding='utf-8').decode('utf-8')


class ODMDoc(XMLRepr):
    """A base ODM document"""

    def __init__(self, xml):
        # Call base class
        XMLRepr.__init__(self, xml)
        r_get = self.root.get

        self.filetype = r_get('FileType')
        self.creationdatetime = r_get('CreationDateTime')
        self.fileoid = r_get("FileOID")
        self.ODMVersion = r_get("ODMVersion")
        self.granularity = r_get("Granularity", None)


class RWSError(ODMDoc):
    """
Extends ODMDoc, inheriting attributes like filetype, creationdatetime etc.

Parses XML of the form::

    <?xml version="1.0" encoding="utf-8"?>
    <ODM xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata"
         FileType="Snapshot"
         CreationDateTime="2013-04-08T10:28:49.578-00:00"
         FileOID="4d13722a-ceb6-4419-a917-b6ad5d0bc30e"
         ODMVersion="1.3"
         mdsol:ErrorDescription="Incorrect login and password combination. [RWS00008]"
         xmlns="http://www.cdisc.org/ns/odm/v1.3" />
    """

    def __init__(self, xml):
        ODMDoc.__init__(self, xml)
        r_get = self.root.get
        self.errordescription = r_get(MEDI_NS + "ErrorDescription")


class RWSErrorResponse(XMLRepr):
    """
Parses messages of the form::

    <Response
        ReferenceNumber="0b47fe86-542f-4070-9e7d-16396a5ef08a"
        InboundODMFileOID="Not Supplied"
        IsTransactionSuccessful="0"
        ReasonCode="RWS00092"
        ErrorClientResponseMessage="CRF version not found">
    </Response>
    """

    def __init__(self, xml):
        # Call base class
        XMLRepr.__init__(self, xml)
        r_get = self.root.get

        self.referencenumber = r_get("ReferenceNumber")
        self.inboundodmfileoid = r_get('InboundODMFileOID')
        self.istransactionsuccessful = r_get('IsTransactionSuccessful') == "1"
        self.reasoncode = r_get("ReasonCode")
        self.errordescription = r_get("ErrorClientResponseMessage")


class RWSResponse(XMLRepr):
    """
Parses messages of the form::

    <Response ReferenceNumber="82e942b0-48e8-4cf4-b299-51e2b6a89a1b"
        InboundODMFileOID=""
        IsTransactionSuccessful="1"
        SuccessStatistics="Rave objects touched: Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0" NewRecords="">
    </Response>
    """

    def __init__(self, xml):
        # Call base class
        XMLRepr.__init__(self, xml)
        r_get = self.root.get

        self.referencenumber = r_get("ReferenceNumber")
        self.inboundodmfileoid = r_get('InboundODMFileOID')
        self.istransactionsuccessful = r_get('IsTransactionSuccessful') == "1"

        self.subjects_touched = 0
        self.folders_touched = 0
        self.forms_touched = 0
        self.fields_touched = 0
        self.loglines_touched = 0

        success_stats = r_get('SuccessStatistics', '')

        # Clinical data post
        if success_stats.startswith('Rave objects touched:'):
            success_stats = success_stats[
                            len('Rave objects touched:') + 1:]  # Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0
            parts = success_stats.split(';')  # [Subjects=0, Folders=0, Forms=0, Fields=0, LogLines=0]
            for part in parts:
                name, value = part.strip().split('=')
                # if value[-1] == ';':
                #     value = value[:-1]
                if name == 'Subjects':
                    self.subjects_touched = int(value)
                elif name == 'Folders':
                    self.folders_touched = int(value)
                elif name == 'Forms':
                    self.forms_touched = int(value)
                elif name == 'Fields':
                    self.fields_touched = int(value)
                elif name == 'LogLines':
                    self.loglines_touched = int(value)
                else:
                    raise KeyError('Unknown Rave Object %s in response %s' % (name, success_stats,))

        # Note: Metadata post has success_stats == 'N/A'
        self.new_records = r_get("NewRecords")


class RWSPostResponse(RWSResponse):
    """
Parses responses from PostODMClinicalData messages with the format::

    <Response ReferenceNumber="82e942b0-48e8-4cf4-b299-51e2b6a89a1b"
              InboundODMFileOID=""
              IsTransactionSuccessful="1"
              SuccessStatistics="Rave objects touched: Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0" NewRecords=""
              SubjectNumberInStudy="1103" SubjectNumberInStudySite="55">
    </Response>
    """

    def __init__(self, xml):
        # Call base class
        RWSResponse.__init__(self, xml)
        r_get = self.root.get
        # These counts may only come as a result of a Clinical data POST
        snis = r_get('SubjectNumberInStudy')
        self.subjects_in_study = int(snis) if snis is not None else None

        sniss = r_get('SubjectNumberInStudySite')
        self.subjects_in_study_site = int(sniss) if sniss is not None else None

        # DraftImported only comes from a MetaData Post
        # In which case successStatistics will be SuccessStatistics="N/A"
        self.draft_imported = r_get("DraftImported")


class RWSPostErrorResponse(RWSResponse):
    """Responses to Clinical data post messages have additional Attributes to normal RWS Response messages::

        <Response
            ReferenceNumber="5b1fa9a3-0cf3-46b6-8304-37c2e3b7d04f"
            InboundODMFileOID="1"
            IsTransactionSuccessful = "0"
            ReasonCode="RWS00024"
            ErrorOriginLocation="/ODM/ClinicalData[1]/SubjectData[1]"
            SuccessStatistics="Rave objects touched: Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0"
            ErrorClientResponseMessage="Subject already exists.">
            </Response>
    """
    def __init__(self, xml):
        RWSResponse.__init__(self, xml)

        # Get additional properties
        r_get = self.root.get
        self.reason_code = r_get('ReasonCode')
        self.error_origin_location = r_get('ErrorOriginLocation')
        self.error_client_response_message = r_get('ErrorClientResponseMessage')


class RWSStudyListItem(object):
    """An item in the RWS Study List response"""

    def __init__(self):
        self.oid = None
        self.studyname = None
        self.protocolname = None
        self.environment = None
        self.projecttype = None

    def isProd(self):
        """Is production if environment is empty"""
        return self.environment == '' and self.projecttype != 'GlobalLibraryVolume'

    @classmethod
    def fromElement(cls, elem):
        """Read properties from an XML Element

         <Study OID="Fixitol(Dev)" mdsol:ProjectType="GlobalLibraryVolume">
            <GlobalVariables>
                  <StudyName>Fixitol (Dev)</StudyName>
                  <StudyDescription/>
                  <ProtocolName>Fixitol</ProtocolName>
            </GlobalVariables>
         </Study>
        """

        self = cls()

        self.oid = elem.get('OID')

        # Not all returned documents have a projecttype (GlobalLibraryVolumes do)
        self.projecttype = elem.get(MEDI_NS + "ProjectType", "Project")

        e_global_variables = elem.find(ODM_NS + 'GlobalVariables')
        self.studyname = e_global_variables.find(ODM_NS + 'StudyName').text
        self.protocolname = e_global_variables.find(ODM_NS + 'ProtocolName').text

        self.environment = getEnvironmentFromNameAndProtocol(self.studyname, self.protocolname)

        return self


# I hate multi-inheritance generally but since this is inheriting from a built-in like list I feel
# less bad about it.
class RWSStudies(list, ODMDoc):
    """
Represents a list of studies. Extends the list class and adds a couple of extra properties::

    <ODM FileType="Snapshot" FileOID="767a1f8b-7b72-4d12-adbe-37d4d62ba75e"
         CreationDateTime="2013-04-08T10:02:17.781-00:00"
         ODMVersion="1.3"
         xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata"
         xmlns:xlink="http://www.w3.org/1999/xlink"
         xmlns="http://www.cdisc.org/ns/odm/v1.3">
         <Study OID="Fixitol(Dev)">
            <GlobalVariables>
                  <StudyName>Fixitol (Dev)</StudyName>
                  <StudyDescription/>
                  <ProtocolName>Fixitol</ProtocolName>
            </GlobalVariables>
         </Study>
         <Study OID="IANTEST(Prod)">
            <GlobalVariables>
                  <StudyName>IANTEST</StudyName>
                  <StudyDescription/>
                  <ProtocolName>IANTEST</ProtocolName>
            </GlobalVariables>
         </Study>
    </ODM>
    """

    def __init__(self, xml):
        # Get basic properties
        ODMDoc.__init__(self, xml)

        for estudy in self.root.findall(ODM_NS + 'Study'):
            self.append(RWSStudyListItem.fromElement(estudy))


class MetaDataVersion(object):
    """
<MetaDataVersion OID="1203" Name="Webservice Outbound"/>
    """

    def __init__(self):
        self.oid = None
        self.name = None

    @classmethod
    def fromElement(cls, elem):
        """Read properties from an XML Element

           <MetaDataVersion OID="1203" Name="Webservice Outbound"/>
        """

        self = cls()

        self.oid = elem.get('OID')
        self.name = elem.get('Name')
        return self


class RWSStudyMetadataVersions(list, ODMDoc, RWSStudyListItem):
    """
Parses responses from MetaDataVersions request::

    <ODM ODMVersion="1.3" Granularity="Metadata" FileType="Snapshot" FileOID="d26b4d33-376d-4037-9747-684411190179" CreationDateTime=" 2013-04-08T01:29:13 " xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata">
        <Study OID="IANTEST">
            <GlobalVariables>
                <StudyName>IANTEST</StudyName>
                <StudyDescription></StudyDescription>
                <ProtocolName>IANTEST</ProtocolName>
            </GlobalVariables>
            <MetaDataVersion OID="1203" Name="Webservice Outbound" />
            <MetaDataVersion OID="1195" Name="JC_Demo_Draft1" />
            <MetaDataVersion OID="1165" Name="Initial" />
        </Study>
    </ODM>
    """

    def __init__(self, xml):
        # Get basic properties
        ODMDoc.__init__(self, xml)

        root = self.root  # from ODMDoc

        e_study = root.find(ODM_NS + 'Study')

        # Quick way to grab the elements here, nasty though
        self.study = RWSStudyListItem.fromElement(e_study)

        for e_version in e_study.findall(ODM_NS + 'MetaDataVersion'):
            self.append(MetaDataVersion.fromElement(e_version))


class RWSSubjectListItem(object):
    """
Parses response of Subject List request::

    <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
     <SubjectData SubjectKey="1111">
        <SiteRef LocationOID="335566"/>
     </SubjectData>
    </ClinicalData>

Optionally ClinicalData may include status::

    <SubjectData SubjectKey="1111" mdsol:Overdue="No"
              mdsol:Touched="Yes"
              mdsol:Empty="No"
              mdsol:Incomplete="No"
              mdsol:NonConformant="No"
              mdsol:RequiresSecondPass="No"
              mdsol:RequiresReconciliation="No"
              mdsol:RequiresVerification="No"
              mdsol:Verified="No"
              mdsol:Frozen="No"
              mdsol:Locked="No"
              mdsol:RequiresReview="No"
              mdsol:PendingReview="No"
              mdsol:Reviewed="No"
              mdsol:RequiresAnswerQuery="No"
              mdsol:RequiresPendingCloseQuery="No"
              mdsol:RequiresCloseQuery="No"
              mdsol:StickyPlaced="No"
              mdsol:Signed="No"
              mdsol:SignatureCurrent="No"
              mdsol:RequiresTranslation="No"
              mdsol:RequiresCoding="No"
              mdsol:RequiresPendingAnswerQuery="No"
              mdsol:RequiresSignature="No"
              mdsol:ReadyForFreeze="No"
              mdsol:ReadyForLock="Yes">

The SubjectKey can be either a Subject ID or a UUID depending on the value of SubjectKeyType::

    <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
     <SubjectData SubjectKey="EC82F1AB-D463-4930-841D-36FC865E63B2" mdsol:SubjectName="1" mdsol:SubjectKeyType="SubjectUUID">
        <SiteRef LocationOID="335566"/>
     </SubjectData>
    </ClinicalData>


    """
    STATUS_PROPERTIES = ["Overdue",
                         "Touched",
                         "Empty",
                         "Incomplete",
                         "NonConformant",
                         "RequiresSecondPass",
                         "RequiresReconciliation",
                         "RequiresVerification",
                         "Verified",
                         "Frozen",
                         "Locked",
                         "RequiresReview",
                         "PendingReview",
                         "Reviewed",
                         "RequiresAnswerQuery",
                         "RequiresPendingCloseQuery",
                         "RequiresCloseQuery",
                         "StickyPlaced",
                         "Signed",
                         "SignatureCurrent",
                         "RequiresTranslation",
                         "RequiresCoding",
                         "RequiresPendingAnswerQuery",
                         "RequiresSignature",
                         "ReadyForFreeze",
                         "ReadyForLock",
                         "SubjectKeyType",
                         "SubjectName"]

    def __init__(self):
        """The ODM message has a ClinicalData element with a single SubjectData and SiteRef elements
           nested within. I collapse into a single object
        """
        self.studyoid = None
        self.metadataversionoid = None
        self.subjectkey = None
        self.subjectkeytype = None
        self.locationoid = None

        self.active = None  # SubjectActive
        self.deleted = None  # Deleted

        # Optional properties, only if status included
        for prop in RWSSubjectListItem.STATUS_PROPERTIES:
            setattr(self, prop.lower(), None)

    @property
    def subject_name(self):
        """
        Get the subject name consistently - if the SubjectKeyType is SubjectUUID
         then the subject name lives in the mdsol:SubjectName attribute

        :return: The Subject ID for the subject
        :rtype
        """
        if self.subjectkeytype and self.subjectkeytype == "SubjectUUID".lower():
            # if the SubjectKeyType is "SubjectUUID", then return the SubjectName
            return self.subjectname
        else:
            return self.subjectkey

    @classmethod
    def fromElement(cls, elem):
        """
        Read properties from an XML Element
        """
        self = cls()
        self.studyoid = elem.get('StudyOID')
        self.metadataversionoid = elem.get('MetaDataVersionOID')
        e_subjectdata = elem.findall(ODM_NS + 'SubjectData')[0]
        self.subjectkey = e_subjectdata.get('SubjectKey')
        e_siteref = e_subjectdata.findall(ODM_NS + 'SiteRef')[0]
        self.locationoid = e_siteref.get('LocationOID')

        decodes = {'yes': True, 'no': False, '': None}
        for prop in RWSSubjectListItem.STATUS_PROPERTIES:
            val = e_subjectdata.get(MEDI_NS + prop, "").lower()
            setattr(self, prop.lower(), decodes.get(val, val))

        # By default we only get back active and non-deleted subjects
        self.active = decodes[e_subjectdata.get(MEDI_NS + "SubjectActive", "yes").lower()]
        self.deleted = decodes[e_subjectdata.get(MEDI_NS + "Deleted", "no").lower()]

        return self


class RWSSubjects(list, ODMDoc):  # I hate multi-inheritance generally.
    """
Represents a list of subjects::

    <ODM FileType="Snapshot"
         FileOID="770f1758-db33-4ab2-af72-38db863734aa"
         CreationDateTime="2013-04-08T14:08:06.875-00:00"
         ODMVersion="1.3">

         <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
            <SubjectData SubjectKey="000001">
               <SiteRef LocationOID="BP001"/>
            </SubjectData>
         </ClinicalData>

         <ClinicalData StudyOID="Fixitol(Dev)" MetaDataVersionOID="1111">
             <SubjectData SubjectKey="1111">
                <SiteRef LocationOID="335566"/>
             </SubjectData>
         </ClinicalData>
    </ODM>
    """

    def __init__(self, xml):
        # Get basic properties
        ODMDoc.__init__(self, xml)

        root = self.root  # from ODMDoc

        for e_clindata in root.findall(ODM_NS + 'ClinicalData'):
            self.append(RWSSubjectListItem.fromElement(e_clindata))



"""
Core request classes that implement calls available in all versions of RWS.

RWSRequest subclasses represent URL endpoints. They can be passed to a RWSConnection get or post method
as appropriate.

TODO: Note that might want to make the Request objects responsible for deciding whether they are get, post, patch etc.

"""
import datetime

from rwslib.rwsobjects import RWSResponse, RWSStudies, RWSStudyMetadataVersions, RWSSubjects, \
    RWSPostResponse
from six.moves.urllib_parse import urlencode


# -----------------------------------------------------------------------------------------------------------------------
# Function utilities I don't want to muddy class heirarchy with

def check_dataset_type(dataset_type):
    """Datasets may only be regular or raw"""
    if dataset_type.lower() not in ['regular', 'raw']:
        raise ValueError("Dataset type not 'regular' or 'raw' is %s" % dataset_type)


def format_date_argument(date_element):
    """
    Take a date as either a datetime.date/datetime or a string and return it as a
     iso8601 formatted value
    :param date_element: passed argument
    :return:
    """
    if not isinstance(date_element, (datetime.datetime, datetime.date,)):
        # TODO:
        if 'T' in date_element:
            _date = datetime.datetime.strptime(date_element, "%Y-%m-%dT%H:%M:%S")
        else:
            _date = datetime.datetime.strptime(date_element, "%Y-%m-%d").date()
    else:
        _date = date_element
    return _date.isoformat()

# -------------------------------------------------------------------------------------------------------
# Utility functions
#
def make_url(*args, **kwargs):
    """Makes a URL from component parts"""
    base = '/'.join(args)
    if kwargs:
        return "%s?%s" % (base, urlencode(kwargs),)
    else:
        return base


class RWSRequest(object):
    """Base class for all RWS Requests"""
    requires_authorization = False
    method = "GET"  # Default

    def result(self, request):
        """Process a result to create a custom output"""
        # By default return text
        return request.text

    def url_path(self):
        """Return url path list"""
        raise NotImplementedError("Override url_path in descendants of RWSRequest") # pragma: no cover

    def args(self):
        """Return additional args here as dict"""
        return {}

    def make_url(self, *args, **kwargs):
        # Note: Including this in the class as a convenience so that you can get all you need from a RWSObject rather
        # than having to import make_url as an additional import.
        return make_url(*args, **kwargs)


# -----------------------------------------------------------------------------------------------------------------------
# Useful subclasses
# -----------------------------------------------------------------------------------------------------------------------

class RWSGetRequest(RWSRequest):
    method = "GET"


class RWSAuthorizedGetRequest(RWSGetRequest):
    requires_authorization = True


class RWSPostRequest(RWSRequest):
    method = "POST"


class RWSAuthorizedPostRequest(RWSPostRequest):
    requires_authorization = True


class QueryOptionGetRequest(RWSAuthorizedGetRequest):
    """Manages requests that have known query string options"""
    KNOWN_QUERY_OPTIONS = []

    def _querystring(self):
        """Get additional keyword arguments"""

        kw = {}
        for key in self.KNOWN_QUERY_OPTIONS:
            val = getattr(self, key)
            if val is not None:
                kw[key] = val
        return kw


# ---------------------------------------------------------------------------------------------------------------------
# Implementations. These are all standards that have existed for a long time.
# ---------------------------------------------------------------------------------------------------------------------

class VersionRequest(RWSGetRequest):
    """Get RWS Version number"""

    def url_path(self):
        return make_url('version')


class BuildVersionRequest(RWSGetRequest):
    """Return the RWS build version number"""

    def url_path(self):
        return make_url('version', 'build')


class DiagnosticsRequest(RWSGetRequest):
    """Return the RWS build version number"""

    def url_path(self):
        return make_url('diagnostics')


class CacheFlushRequest(RWSAuthorizedGetRequest):
    """Calls RWS cache-flush"""

    def url_path(self):
        return make_url('webservice.aspx?CacheFlush')

    def result(self, request):
        """Return RWSResponse object for success"""
        return RWSResponse(request.text)


class ClinicalStudiesRequest(RWSAuthorizedGetRequest):
    """Return the list of clinical studies as a RWSStudies object.
       Clinical studies are the studies that you have access to as an EDC user.
    """

    def url_path(self):
        return make_url('studies')

    def result(self, request):
        return RWSStudies(request.text)


# ----------------------------------------------------------------------------------------------------------------------
# Base request classes for study versions (could also be used for drafts if ever implemented by RWS)
# ----------------------------------------------------------------------------------------------------------------------

class VersionRequestBase(RWSAuthorizedGetRequest):
    """Base class for study and library metadata version requests"""

    def __init__(self, project_name, oid):
        self.project_name = project_name
        self._oid = None  # Oid is a version OID, an integer identifier
        self.oid = oid

    @property
    def oid(self):
        return self._oid

    @oid.setter
    def oid(self, value):
        try:
            int(value)
        except ValueError:
            raise ValueError('oid must be an integer')
        self._oid = value

    def result(self, request):
        return request.text


# -----------------------------------------------------------------------------------------------------------------------
# Related to Architect Studies and their drafts/versions
# -----------------------------------------------------------------------------------------------------------------------

class MetadataStudiesRequest(RWSAuthorizedGetRequest):
    """Return the list of metadata studies as a RWSStudies object.
       metadata_studies are the list of studies that you have access to as an
       Architect user.
    """

    def url_path(self):
        return make_url('metadata', 'studies')

    def result(self, request):
        return RWSStudies(request.text)


class StudyDraftsRequest(RWSAuthorizedGetRequest):
    """Return the list of study drafts"""

    def __init__(self, project_name):
        self.project_name = project_name

    def url_path(self):
        return make_url('metadata', 'studies', self.project_name, 'drafts')

    def result(self, request):
        return RWSStudyMetadataVersions(request.text)


class StudyVersionsRequest(RWSAuthorizedGetRequest):
    """Return the list of study versions"""

    def __init__(self, project_name):
        self.project_name = project_name

    def url_path(self):
        return make_url('metadata', 'studies', self.project_name, 'versions')

    def result(self, request):
        return RWSStudyMetadataVersions(request.text)


class StudyVersionRequest(VersionRequestBase):
    """Return a study version as a string"""

    def url_path(self):
        return make_url('metadata', 'studies', self.project_name, 'versions', str(self._oid))


# NOTE: There is no StudyDraftRequest, this is something of an omission since you can list them...


# -------------------------------------------------------------------------------------------------
# Related to Architect Global Libraries and their drafts/versions
# -------------------------------------------------------------------------------------------------

class GlobalLibrariesRequest(RWSAuthorizedGetRequest):
    """Return the list of global libraries as a RWSStudies object.
       metadata_libraries are the list of libraries that you have access to as an
       Architect Global Library Volume user
    """

    def url_path(self):
        return make_url('metadata', 'libraries')

    def result(self, request):
        return RWSStudies(request.text)


class GlobalLibraryDraftsRequest(RWSAuthorizedGetRequest):
    """Return the list of global library drafts"""

    def __init__(self, project_name):
        self.project_name = project_name

    def url_path(self):
        return make_url('metadata', 'libraries', self.project_name, 'drafts')

    def result(self, request):
        return RWSStudyMetadataVersions(request.text)


class GlobalLibraryVersionsRequest(RWSAuthorizedGetRequest):
    """Return the list of global library versions"""

    def __init__(self, project_name):
        self.project_name = project_name

    def url_path(self):
        return make_url('metadata', 'libraries', self.project_name, 'versions')

    def result(self, request):
        return RWSStudyMetadataVersions(request.text)


class GlobalLibraryVersionRequest(VersionRequestBase):
    """Return a global library version as a string"""

    def url_path(self):
        return make_url('metadata', 'libraries', self.project_name, 'versions', str(self._oid))


class PostMetadataRequest(RWSAuthorizedPostRequest):
    """Post an ODM data transaction to Rave, get back an RWSResponse object"""

    def __init__(self, project_name, data, headers={'Content-type': "text/xml"}):
        self.project_name = project_name
        self.data = data
        self.headers = headers

    def args(self):
        """Return additional args here as dict (only for post data requests)"""
        kw = {'data': self.data,
              'headers': self.headers}
        return kw

    def url_path(self):
        return make_url('metadata', 'studies', self.project_name, 'drafts')

    def result(self, request):
        return RWSPostResponse(request.text)


# -------------------------------------------------------------------------------------------------
# Subject related
# -------------------------------------------------------------------------------------------------

class StudySubjectsRequest(RWSAuthorizedGetRequest):
    """
    Return the list of study subjects, defaults to the PROD environment
    """
    SUBJECT_KEY_TYPES = ["SubjectName", "SubjectUUID"]
    INCLUDE_OPTIONS = ['inactive', 'deleted', 'inactiveAndDeleted']

    def __init__(self, project_name, environment_name, status=False, include=None,
                 subject_key_type='SubjectName'):
        """
        If status == True then ?status=all
        if include then include parameter is also added to query string
        """

        self.project_name = project_name
        self.environment_name = environment_name
        self.status = status
        self.include = None
        self.subject_key_type = subject_key_type
        # make sure the value for SubjectKeyType makes sense.
        if self.subject_key_type not in self.SUBJECT_KEY_TYPES:
            raise ValueError(
                "SubjectKeyType {} is not a valid value".format(self.subject_key_type))

        if include is not None:
            if include not in self.INCLUDE_OPTIONS:
                raise ValueError(
                    'If provided, included must be one of %s' % ','.join(self.INCLUDE_OPTIONS))
        self.include = include

    def _querystring(self):
        """Additional keyword arguments"""
        kw = {}
        if self.status:
            kw['status'] = 'all'

        if self.include is not None:
            kw['include'] = self.include
        if self.subject_key_type != 'SubjectName':
            kw['subjectKeyType'] = self.subject_key_type
        return kw

    def studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name,)

    def url_path(self):
        return make_url('studies', self.studyname_environment(), 'subjects', **self._querystring())

    def result(self, request):
        return RWSSubjects(request.text)


class PostDataRequest(RWSAuthorizedPostRequest):
    """Post an ODM data transaction to Rave, get back an RWSResponse object"""

    def __init__(self, data, headers={'Content-type': "text/xml"}):
        self.data = data
        self.headers = headers

    def args(self):
        """Return additional args here as dict (only for post data requests)"""
        kw = {'data': self.data,
              'headers': self.headers}
        return kw

    def url_path(self):
        return make_url("webservice.aspx?PostODMClinicalData")

    def result(self, request):
        return RWSPostResponse(request.text)


# -------------------------------------------------------------------------------------------------
# ODM Clinical Data Datasets
# -------------------------------------------------------------------------------------------------


class ODMDatasetBase(RWSAuthorizedGetRequest):
    KNOWN_QUERY_OPTIONS = ['versionitem', 'rawsuffix', 'codelistsuffix',
                           'decodesuffix', 'stdsuffix', 'start']

    def checkParams(self):
        check_dataset_type(self.dataset_type)
        # https://bitbucket.org/micktwomey/pyiso8601

    def _querystring(self):
        """Additional keyword arguments"""

        kw = {}

        for key in self.KNOWN_QUERY_OPTIONS:
            val = getattr(self, key)
            if val is not None:
                if key == 'start':
                    # special case the date formatting
                    # this will raise a ValueError
                    # if the date format is a string and not one of
                    # YYYY-mm-ddTHH:MM:SS or YYYY-mm-dd
                    kw[key] = format_date_argument(val)
                else:
                    kw[key] = val
        return kw

    def _studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name,)


class StudyDatasetRequest(ODMDatasetBase):
    """Return the text of the full datasets listing as an ODM string."""

    def __init__(self, project_name, environment_name, dataset_type='regular',
                 start=None, rawsuffix=None, formoid=None,
                 versionitem=None, codelistsuffix=None, decodesuffix=None, stdsuffix=None):
        self.project_name = project_name
        self.environment_name = environment_name

        self.dataset_type = dataset_type.lower()
        self.formoid = formoid

        self.rawsuffix = rawsuffix
        self.start = start
        self.versionitem = versionitem
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix
        self.stdsuffix = stdsuffix

        self.checkParams()

    def url_path(self):
        args = ['studies', self._studyname_environment(), 'datasets', self.dataset_type]
        if self.formoid is not None:
            args.append(self.formoid)

        return make_url(*args, **self._querystring())


class VersionDatasetRequest(ODMDatasetBase):
    """
    Return the text of the full datasets for a version as an ODM string.
    By supplying formoid, will be filtered to just that formoid data
    """

    def __init__(self, project_name, environment_name, version_oid, dataset_type='regular',
                 start=None, rawsuffix=None, formoid=None,
                 versionitem=None, codelistsuffix=None, decodesuffix=None, stdsuffix=None):
        self.project_name = project_name
        self.environment_name = environment_name
        self.version_oid = version_oid

        self.dataset_type = dataset_type.lower()
        self.formoid = formoid

        self.rawsuffix = rawsuffix
        self.start = start
        self.versionitem = versionitem
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix
        self.stdsuffix = stdsuffix

        self.checkParams()

    def url_path(self):
        args = ['studies', self._studyname_environment(),
                'versions', str(self.version_oid),
                'datasets', self.dataset_type]
        if self.formoid is not None:
            args.append(self.formoid)

        return make_url(*args, **self._querystring())


class SubjectDatasetRequest(ODMDatasetBase):
    """
    Return the text of the full datasets for a version as an ODM string.
    By supplying formoid, will be filtered to just that formoid data
    """

    def __init__(self, project_name,
                 environment_name,
                 subjectkey,
                 dataset_type='regular', start=None, rawsuffix=None, formoid=None,
                 versionitem=None, codelistsuffix=None, decodesuffix=None, stdsuffix=None):
        self.project_name = project_name
        self.environment_name = environment_name
        self.subjectkey = subjectkey

        self.dataset_type = dataset_type.lower()
        self.formoid = formoid

        self.rawsuffix = rawsuffix
        self.start = start
        self.versionitem = versionitem
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix
        self.stdsuffix = stdsuffix

        self.checkParams()

    def url_path(self):
        args = ['studies', self._studyname_environment(),
                'subjects', str(self.subjectkey),
                'datasets', self.dataset_type]
        if self.formoid is not None:
            args.append(self.formoid)

        return make_url(*args, **self._querystring())



"""
Requests related to the ODM Adapter

http://rws-webhelp.s3.amazonaws.com/WebHelp_ENG/solutions/clinical_data_audits/index.html#odm-adapter
"""
from . import RWSAuthorizedGetRequest, QueryOptionGetRequest

class AuditRecordsRequest(QueryOptionGetRequest):
    """Clinical Audit Records Dataset"""
    KNOWN_QUERY_OPTIONS = ['studyoid','startid','per_page']

    def __init__(self, project_name, environment_name, startid=1, per_page=100):
        self.project_name = project_name
        self.environment_name = environment_name
        self.startid = startid
        self.per_page = per_page

    @property
    def studyoid(self):
        """Create studyoid from project name and environment"""
        return "%s(%s)" % (self.project_name, self.environment_name,)

    def url_path(self):
        """Return url path list"""
        return self.make_url('datasets', 'ClinicalAuditRecords.odm', **self._querystring())

class VersionFoldersRequest(QueryOptionGetRequest):
    """Identify all folders in use in study"""
    KNOWN_QUERY_OPTIONS = ['studyoid']

    def __init__(self, project_name, environment_name):
        self.project_name = project_name
        self.environment_name = environment_name

    @property
    def studyoid(self):
        """Create studyoid from project name and environment"""
        return "%s(%s)" % (self.project_name, self.environment_name,)

    def _dataset_name(self):
        return 'VersionFolders.odm'

    def url_path(self):
        """Return url path list"""
        return self.make_url('datasets', self._dataset_name(), **self._querystring())


class SitesMetadataRequest(RWSAuthorizedGetRequest):
    """List all sites in a study along with their StudyVersions"""

    def __init__(self, project_name=None, environment_name=None):
        self.project_name = project_name
        self.environment_name = environment_name

        if project_name is not None:
            if environment_name in [None,'']:
                raise AttributeError("environment_name cannot be empty if project_name is set")

        if environment_name is not None:
            if project_name in [None,'']:
                raise AttributeError("project_name cannot be empty if environment_name is set")

    @property
    def studyoid(self):
        """Create studyoid from project name and environment"""
        return "%s(%s)" % (self.project_name, self.environment_name,)

    def _querystring(self):
        """Additional keyword arguments if filtered to project_name and environment_name"""
        if self.project_name is not None:
            return {'studyoid': self.studyoid }
        return {}

    def url_path(self):
        """Return url path list"""
        return self.make_url('datasets', 'Sites.odm', **self._querystring())


class UsersRequest(RWSAuthorizedGetRequest):
    """Return list of users for study (can be filtered by location)"""
    def __init__(self, project_name, environment_name, location_oid=None):
        self.project_name = project_name
        self.environment_name = environment_name
        self.location_oid = location_oid

    @property
    def studyoid(self):
        """Create studyoid from project name and environment"""
        return "%s(%s)" % (self.project_name, self.environment_name,)

    def _querystring(self):
        """Additional keyword arguments"""

        kw = {'studyoid': self.studyoid}
        if self.location_oid is not None:
            kw['locationoid'] = self.location_oid
        return kw

    def url_path(self):
        """Return url path list"""
        return self.make_url('datasets', 'Users.odm', **self._querystring())


class SignatureDefinitionsRequest(RWSAuthorizedGetRequest):
    """Return signature definitions for all versions of the study"""

    def __init__(self, project_name):
        self.project_name = project_name

    def _querystring(self):
        """Additional keyword arguments"""

        kw = {'studyid': self.project_name}
        return kw

    def url_path(self):
        """Return url path list"""
        return self.make_url('datasets', 'Signatures.odm', **self._querystring())



"""
Views related to clinical views and their metadata, Biostats Gateway/Adapter

http://rws-webhelp.s3.amazonaws.com/WebHelp_ENG/solutions/01_biostat_adapter.html#biostat-adapter

"""

from . import RWSAuthorizedGetRequest, QueryOptionGetRequest, check_dataset_type

#-----------------------------------------------------------------------------------------------------------------------
# Utilities

DATASET_FORMATS = {'csv': '.csv', 'xml': ''}


def check_dataset_format(ds_format):
    """Ensure dataset format is XML or CSV"""
    if ds_format.lower() not in DATASET_FORMATS.keys():
        raise ValueError("dataset_format is expected to be one of %s. '%s' is not valid" % (
        ', '.join(DATASET_FORMATS.keys()), ds_format,))


def dataset_format_to_extension(ds_format):
    try:
        return DATASET_FORMATS[ds_format]
    except KeyError:
        raise ValueError("dataset_format is expected to be one of %s. '%s' is not valid" % (
        ', '.join(DATASET_FORMATS.keys()), ds_format,))


#-----------------------------------------------------------------------------------------------------------------------
# Classes


class CVMetaDataRequest(QueryOptionGetRequest):
    """Return Clinical View Metadata as ODM string"""
    KNOWN_QUERY_OPTIONS = ['versionitem', 'rawsuffix', 'codelistsuffix', 'decodesuffix']

    def __init__(self, project_name, environment_name, versionitem=None, rawsuffix=None, codelistsuffix=None,
                 decodesuffix=None):
        self.project_name = project_name
        self.environment_name = environment_name

        self.versionitem = versionitem
        self.rawsuffix = rawsuffix
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix

    def studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name, )


    def url_path(self):
        return self.make_url('studies', self.studyname_environment(), 'datasets', 'metadata', 'regular',
                             **self._querystring())


class FormDataRequest(QueryOptionGetRequest):
    """Return CV Form Data as CSV or XML"""
    KNOWN_QUERY_OPTIONS = ['start']

    def __init__(self, project_name, environment_name, dataset_type, form_oid, start=None, dataset_format="csv"):
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format

        self.project_name = project_name
        self.environment_name = environment_name

        self.dataset_type = dataset_type
        check_dataset_type(self.dataset_type)

        self.form_oid = form_oid

        #TODO: Check start is iso8601 date/time
        self.start = start

    def studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name, )

    def _dataset_name(self):
        return '%s%s' % (self.form_oid, dataset_format_to_extension(self.dataset_format),)

    def url_path(self):
        return self.make_url('studies', self.studyname_environment(), 'datasets', self.dataset_type,
                             self._dataset_name(), **self._querystring())


class MetaDataRequest(RWSAuthorizedGetRequest):
    """Return Metadata for Clinical Views in CSV or XML fornat"""

    def __init__(self, dataset_format="csv"):
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format

    def _dataset_name(self):
        return 'ClinicalViewMetadata%s' % dataset_format_to_extension(self.dataset_format)

    def url_path(self):
        return self.make_url('datasets', self._dataset_name())


class ProjectMetaDataRequest(RWSAuthorizedGetRequest):
    """Return Metadata for Clinical Views in CSV or XML fornat for a Project"""

    def __init__(self, project_name, dataset_format="csv"):
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format.lower()
        self.project_name = project_name

    def _dataset_name(self):
        return 'ClinicalViewMetadata%s' % dataset_format_to_extension(self.dataset_format)

    def url_path(self):
        return self.make_url('datasets', self._dataset_name(), **{'ProjectName': self.project_name})


class ViewMetaDataRequest(RWSAuthorizedGetRequest):
    """Return Metadata for Clinical Views in CSV fornat for a single View"""

    def __init__(self, view_name, dataset_format="csv"):
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format.lower()
        self.view_name = view_name

    def _dataset_name(self):
        return 'ClinicalViewMetadata%s' % dataset_format_to_extension(self.dataset_format)

    def url_path(self):
        return self.make_url('datasets', self._dataset_name(), **{'ViewName': self.view_name})


class CommentDataRequest(RWSAuthorizedGetRequest):
    """Return Comments from Rave as CSV or XML"""

    def __init__(self, project_name, environment_name, dataset_format="csv"):
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format.lower()
        self.project_name = project_name
        self.environment_name = environment_name

    def studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name, )

    def _dataset_name(self):
        return 'SDTMComments%s' % dataset_format_to_extension(self.dataset_format)

    def url_path(self):
        return self.make_url('datasets', self._dataset_name(), **{'studyid': self.studyname_environment()})


class ProtocolDeviationsRequest(CommentDataRequest):
    """Retrieve Protocol Deviation Information from Rave"""

    def _dataset_name(self):
        return 'SDTMProtocolDeviations%s' % dataset_format_to_extension(self.dataset_format)

    def url_path(self):
        return self.make_url('datasets', self._dataset_name(), **{'studyid': self.studyname_environment()})


class DataDictionariesRequest(CommentDataRequest):
    """Retrieve Data Dictionaries from Rave"""

    def _dataset_name(self):
        return 'SDTMDataDictionaries%s' % dataset_format_to_extension(self.dataset_format)

    def url_path(self):
        return self.make_url('datasets', self._dataset_name(), **{'studyid': self.studyname_environment()})



#!/usr/bin/env python

import os
import sys
import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

packages = [
    'rwslib',
    'rwslib.rws_requests',
]

rwsinit = open('rwslib/__init__.py').read()
author = re.search("__author__ = '([^']+)'", rwsinit).group(1)
version = re.search("__version__ = '([^']+)'", rwsinit).group(1)

setup(
    name='rwslib',
    version=version,
    description='Rave web services for Python',
    long_description=open('README.md').read(),
    author=author,
    author_email="isparks@mdsol.com",
    packages=packages,
    package_dir={'rwslib': 'rwslib'},
    include_package_data=True,
    install_requires=['requests', 'lxml', 'httpretty', 'six'],
    license=open('LICENSE.txt').read(),
    zip_safe=False,
    test_suite='rwslib.tests.all_tests',
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
    ),
)
